"""Module for type hint annotations."""


from fondat.validation import validate_arguments
from typing import Any


class Annotation:
    """Base class for annotations."""

    __slots__ = {"value"}

    def __repr__(self):
        return f"{type(self).__name__}({self.value!r})"

    def __str__(self):
        return str(self.value)

    def __eq__(self, other: Any):
        return type(self) == type(other) and self.value == other.value

    def __hash__(self):
        try:
            return hash((self.__class__, self.value))
        except:
            return super().__hash__()


class Deprecated(Annotation):
    """Type annotation to indicate a value is deprecated."""

    @validate_arguments
    def __init__(self, value: bool):
        self.value = value


class Description(Annotation):
    """Type annotation to provide a textual description."""

    @validate_arguments
    def __init__(self, value: str):
        self.value = value


class Example(Annotation):
    """Type annotation to provide an example value."""

    def __init__(self, value: Any):
        self.value = value


class Format(Annotation):
    """Type annotation to express string format."""

    @validate_arguments
    def __init__(self, value: str):
        self.value = value


class ReadOnly(Annotation):
    """Type annotation to indicate a value is read-only."""

    @validate_arguments
    def __init__(self, value: bool):
        self.value = value


Password = Format("password")



"""Module to manage types and type hints."""

import contextvars
import dataclasses
import functools
import types
import typing

from collections.abc import Iterable, Mapping
from contextlib import contextmanager
from types import NoneType, UnionType
from typing import Any, TypeVar, get_args, get_origin


_typevars = contextvars.ContextVar("_fondat_typevars")


def affix_type_hints(
    obj: Any = None,
    *,
    globalns: Mapping[str, Any] | None = None,
    localns: Mapping[str, Any] | None = None,
    attrs: bool = True,
):
    """
    Affixes an object's type hints to the object by materializing evaluated string type hints
    into the type's __annotations__ attribute.

    This function exists due to PEP 563, in which annotations can be stored as strings, and are
    only evaluated when typing.get_type_hints is called; this will be the expected behavior of
    annotations in Python 3.11. The work in PEP 649, if accepted, will likely eliminate the
    need to affix type hints.

    This function can be applied as a decorator to a class or function.

    Parameters:
    • obj: module, class, method or function
    • globalns: global namespace to evaluate type hints
    • localns: local namespace to evaluate type hints
    • attrs: affix all of object's attribute type hints

    Type hints are affixed by first resolving through typing.get_type_hints, then by storing
    the result in the object's __annotations__ attribute.

    If the object is a class, this function will affix annotations from all superclasses into
    the object annotations.

    Affixation provides the following benefits (under PEP 563):
    • time and scope of annotation evaluation is under the control of the caller
    • annotations are not re-evaluated for every call to typing.get_type_hints
    """

    if obj is None:
        return functools.partial(
            affix_type_hints, globalns=globalns, localns=localns, attrs=attrs
        )

    obj = strip_annotations(obj)

    if getattr(obj, "__annotations__", None):
        obj.__annotations__ = typing.get_type_hints(obj, globalns, localns, include_extras=True)

    if dataclasses.is_dataclass(obj):
        for field in dataclasses.fields(obj):
            if type := obj.__annotations__.get(field.name):
                field.type = type

    if attrs:
        for name in dir(obj):
            if not name.startswith("__") and not name.endswith("__"):
                affix_type_hints(
                    getattr(obj, name), globalns=globalns, localns=localns, attrs=False
                )

    return obj


def split_annotations(type_hint: Any) -> tuple[Any, tuple[Any, ...]]:
    """Return a tuple separating the python type and annotations."""
    if typing.get_origin(type_hint) is not typing.Annotated:
        return type_hint, ()
    args = typing.get_args(type_hint)
    return args[0], args[1:]


def strip_annotations(type_hint: Any) -> tuple[Any, tuple[Any, ...]]:
    """Return type hint with typing annotations stripped."""
    if typing.get_origin(type_hint) is not typing.Annotated:
        return type_hint
    return typing.get_args(type_hint)[0]


def is_optional(type_hint: Any) -> bool:
    """
    Return if the specified type is optional.

    A type is optional if its type hint matches any of the following:
    • None
    • Optional[...]
    • Union[..., None]
    • ... | None
    """
    python_type = strip_annotations(type_hint)
    if not typing.get_origin(python_type) in {types.UnionType, typing.Union}:
        return python_type is NoneType
    for arg in typing.get_args(python_type):
        if is_optional(arg):
            return True
    return False


def strip_optional(type_hint):
    """Return a union type with optionality stripped."""
    python_type, annotations = split_annotations(type_hint)
    origin = typing.get_origin(python_type)
    if origin not in {types.UnionType, typing.Union}:
        return type_hint
    args = (strip_optional(arg) for arg in typing.get_args(python_type) if arg is not NoneType)
    python_type = union_type(args)
    if not annotations:
        return python_type
    return typing.Annotated[tuple([python_type, *annotations])]


def is_subclass(cls: Any, classinfo: type | tuple[type, ...] | UnionType) -> bool:
    """A more forgiving issubclass."""
    try:
        return issubclass(cls, classinfo)
    except:
        return False


def is_instance(obj: Any, classinfo: type | tuple[type, ...] | UnionType) -> bool:
    """A more forgiving isinstance."""
    try:
        return isinstance(obj, classinfo)
    except:
        return False


def literal_values(literal_type_hint) -> tuple[Any]:
    """Return a set of all values in a Literal type."""
    raw_type = strip_annotations(literal_type_hint)
    return typing.get_args(raw_type)


def union_type(type_hints: Iterable[Any]) -> types.UnionType:
    """Construct a union type from an iterable of types."""
    types = iter(type_hints)
    try:
        result = types.__next__()
    except StopIteration:
        return NoneType
    while True:
        try:
            result |= types.__next__()
        except StopIteration:
            break
    return result


@contextmanager
def capture_typevars(alias: Any):
    """
    Return a context manager that captures type variable substitions in a provided generic
    alias, to be resolved through calls to `resolve_typevar` within the context. This allows
    classes to contain generic types, and for those types to be resolved at runtime. If
    the passed value is not a generic alias, nothing is captured.
    """
    typevars = None
    if args := get_args(alias):
        if params := getattr(get_origin(alias), "__parameters__", None):
            typevars = {p: a for p, a in zip(params, args) if isinstance(p, TypeVar)}
    token = None
    if typevars:
        if outer := _typevars.get(None):  # nested generic alias
            typevars = outer | typevars
        token = _typevars.set(typevars)
    yield
    if token:
        _typevars.reset(token)


def resolve_typevar(typevar: TypeVar) -> Any:
    """
    Resolve a captured type variable substituion from a containing generic class instance.
    For more information, see `capture_typevars` function.
    """
    result = typevar
    while isinstance(result, TypeVar):
        result = _typevars.get({}).get(result) or Any
    return result



"""Resource error module."""

import http

from collections.abc import Iterator


class Error(Exception):
    """
    Base class for resource errors.

    All error classes must include the following attributes:
    • status: HTTP status code (int)
    • phrase: HTTP reason phrase
    """


class ClientError(Error):
    """
    Base class for client resource errors.
    """


class ServerError(Error):
    """
    Base class for server resource errors.
    """


class _Errors:
    """
    Encapsulates resource error exception classes. Errors are dynamically generated from
    errors in the http.HTTPStatus enum.

    Errors can be accessed by HTTP status or name.
    Example: fondat.error.errors[404] == fondat.error.errors.NotFoundError
    """

    def __init__(self):
        self._names = {}
        self._codes = {}
        for status in (s for s in http.HTTPStatus if 400 <= s.value <= 599):
            name = "".join(
                w.title() if w not in {"HTTP", "URI"} else w for w in status.name.split("_")
            )
            if not name.endswith("Error"):
                name += "Error"
            error = type(
                name,
                (ClientError if 400 <= status.value <= 499 else ServerError,),
                {
                    "status": status.value,
                    "phrase": status.phrase,
                    "__doc__": f"{status.description or status.phrase.capitalize()}.",
                },
            )
            self._names[name] = error
            self._codes[status.value] = error

    def get(self, code: int, default=None) -> Error:
        """Return error for code."""
        return self._codes.get(code, default)

    def __getitem__(self, code: int) -> Error:
        return self._codes[code]

    def __getattr__(self, name: str) -> Error:
        if error := self._names.get(name):
            return error
        raise AttributeError

    def __iter__(self) -> Iterator[Error]:
        return iter(self._codes.values())


errors = _Errors()


# commonly used errors
BadRequestError: ClientError = errors.BadRequestError
ConflictError: ClientError = errors.ConflictError
ForbiddenError: ClientError = errors.ForbiddenError
InternalServerError: ServerError = errors.InternalServerError
MethodNotAllowedError: ClientError = errors.MethodNotAllowedError
NotFoundError: ClientError = errors.NotFoundError
UnauthorizedError: ClientError = errors.UnauthorizedError



"""Module to support lazy evaluation of values."""

import importlib
import logging
import threading

from collections.abc import Callable, Mapping, MutableMapping
from typing import Any


_logger = logging.getLogger(__name__)


class LazyMap(MutableMapping):
    """
    A mapping of key-value pairs, in which a value can be lazily initialized at the time of
    first access. This is useful to allow resources to access other resources, preventing
    circular imports.

    Parameters:
    • init: mapping of key-value pairs to initialize the lazy map

    To lazily initialize a value, set it to a no-argument function callback that has been
    decorated with the @lazy decorator. When the value is then first accessed, the callback
    will be called to initialize the value. The resulting value will then be stored in the
    mapping.
    """

    def __init__(self, init: Mapping | None = None):
        super().__init__()
        self._lock = threading.Lock()
        self._store = {}
        if init is not None:
            for key, value in init.items():
                self[key] = value

    def __getitem__(self, key):
        value = self._store[key]
        if is_lazy(value):
            with self._lock:
                value = self._store[key]
                if is_lazy(value):  # prevent race
                    self._store[key] = value = value()
        return value

    def __setitem__(self, key, value):
        self._store[key] = value

    def __delitem__(self, key):
        del self._store[key]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def __contains__(self, item):
        return item in self._store


def lazy(function: Any) -> Any:
    """Decorate a function to tag it as a lazy initializer."""
    if not callable(function):
        raise TypeError("function must be callable")
    setattr(function, "_fondat_lazy", True)
    return function


def lazy_import(module_name: str) -> Callable:
    """
    Return a lazy callback function that imports a module and returns it.

    Parameters:
    • module_name: the name of the module to import
    """

    def callback():
        return importlib.import_module(module_name)

    return lazy(callback)


def lazy_import_attr(module_name: str, attr_name: str) -> Callable:
    """
    Return a lazy callback function that imports a module and returns an attribute from it.

    Parameters:
    • module_name: the name of the module to import
    • attr_name: the name of the attribute in the module to be returned
    """

    def callback():
        return getattr(importlib.import_module(module_name), attr_name)

    return lazy(callback)


def is_lazy(obj: Any) -> bool:
    """Return True if an object is a callable is tagged as a lazy callback function."""
    return callable(obj) and getattr(obj, "_fondat_lazy", None)



"""
Module to manage execution context stacks.

The execution context stack allows values to be "built up" as a request is handled and
operations are performed. It can provide information about the origin of a request, identify
principal(s) making a request, and include access tokens or any other information that can be
used when making authorization decisions.

The execution context stack is stored in context-local state; therefore sepeate stacks are
maintained for different threads and asynchrous tasks.
"""

import contextvars
import datetime
import uuid

from typing import Any, Generator


_stack = contextvars.ContextVar("_fondat_stack")


class _Element:
    """
    A context stack element.

    The context stack is a linked list of elements; each stack element contains a value and a
    reference to the previously pushed element below it on the stack. Because stacks are linked
    lists, they can be safely forked when asynchronous tasks are performed.

    Each element is iterable; it will iterate over the values of the entire stack, beginning
    with its value, then the previously pushed element's value below it, and so on. In this
    manner, the element at the top of a stack represents the entire stack.

    Parameters:
    • value: the value this stack element contains
    • prev: the element below this elemment on the stack, or None if this is the first element
    """

    __slots__ = {"_value", "_prev", "_len"}

    def __init__(self, value, prev=None):
        self._value = value
        self._prev = prev
        self._len = prev._len + 1 if prev else 1

    def __iter__(self):
        class _iter:
            __slots__ = {"_ptr"}

            def __init__(self, ptr):
                self._ptr = ptr

            def __iter__(self):
                return self

            def __next__(self):
                if self._ptr is None:
                    raise StopIteration
                result = self._ptr._value
                self._ptr = self._ptr._prev
                return result

        return _iter(self)

    def __len__(self):
        return self._len


class StackContextManager:
    """
    A context manager returned from pushing a value on the context stack, to automatically pop
    a value from the stack upon exit.
    """

    __slots__ = {"_token"}

    def __init__(self, token):
        self._token = token

    def pop(self):
        """
        Manually pop the pushed value from the stack. Use with caution. The preferred use of
        this context manager is to use "with" keyword to pop from the stack upon exit.
        """
        if self._token:
            _stack.reset(self._token)
        self._token = None

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.pop()


def push(*args, **kwargs) -> StackContextManager:
    """
    Push a new value onto the execution context stack, and return a context manager that will
    pop the value from the stack upon exit.

    A context value is a mapping of key-value pairs. It is expressed as:
    • push(mapping): element is initialized from a mapping object's key-value pairs
    • push(**kwargs): element is initialized with name-value pairs in keyword arguments

    A pushed value must contain a "context" value to express the type of context being pushed
    onto the stack; context keys and "context" value beginning with "fondat." are reserved.

    If no context-local stack exists, then pushing a value causes a new stack to be created
    with an initial "fondat.root" element, which will contain a unique identifier "id" and
    timestamp "time".
    """
    value = dict(*args, **kwargs)
    if "context" not in value:
        raise ValueError('pushed context must have a "context" item')
    stack = _stack.get(None)
    if not stack:
        stack = _Element(
            dict(
                context="fondat.root",
                id=uuid.uuid4(),
                time=datetime.datetime.now(tz=datetime.timezone.utc),
            )
        )
    token = _stack.set(_Element(value, stack))
    return StackContextManager(token)


def find(*args, **kwargs) -> Generator[Any, None, None]:
    """
    Return a generator that yields elements on the context stack that match the specified keys
    and values. Elements are returned in the order of most recently pushed to least recently
    pushed.

    The elements to match can be expressed as follows:
    • find(mapping): match is expressed as a mapping object's key-value pairs
    • find(**kwargs): match is expressed with name-value pairs in keyword arguments

    Supplying no parameters will yield all elements on the stack.
    """
    test = dict(*args, **kwargs).items() or None
    return (value for value in _stack.get(()) if test is None or test <= value.items())


def first(*args, **kwargs) -> Any:
    """
    Return the least recently pushed element on the contact stack that matches the specified
    keys and values, or None if no such element exists.

    The element to match for can be expressed as follows:
    • first(mapping): match is expressed as a mapping object's key-value pairs
    • first(**kwargs): match is expressed with name-value pairs in keyword arguments
    """
    result = None
    for result in find(*args, **kwargs):
        pass
    return result


def last(*args, **kwargs) -> Any:
    """
    Return the most recently pushed element on the contact stack that matches the specified
    keys and values, or None if no such element exists.

    The element to match for can be expressed as follows:
    • last(mapping): match is expressed as a mapping object's key-value pairs
    • last(**kwargs): match is expressed with name-value pairs in keyword arguments
    """
    return next(iter(find(*args, **kwargs)), None)



"""Dataclass module."""

import dataclasses
import functools

from collections.abc import Iterable, Mapping
from dataclasses import is_dataclass
from fondat.annotation import Password
from fondat.types import is_optional, is_subclass, split_annotations, strip_optional
from typing import Any, TypedDict, get_type_hints


def _datacls_init(dc: Any):
    class MISSING:
        pass

    fields = {field.name: field for field in dataclasses.fields(dc) if field.init}

    def __init__(self, **kwargs):

        hints = get_type_hints(self, include_extras=True)

        for name in kwargs:
            if name not in fields:
                raise TypeError(f"unexpected keyword argument: '{name}'")

        for field in fields.values():
            value = kwargs.get(field.name, MISSING)
            if value is MISSING:
                if field.default_factory is not dataclasses.MISSING:
                    value = field.default_factory()
                elif field.default is not dataclasses.MISSING:
                    value = field.default
                elif is_optional(hints[field.name]):
                    value = None
                else:
                    raise TypeError(f"missing required keyword argument: '{field.name}'")
            setattr(self, field.name, value)

        post_init = getattr(self, "__post_init__", MISSING)
        if post_init is not MISSING:
            post_init()

    return __init__


def datacls(cls: type, init: bool = True, **kwargs) -> type:
    """
    Decorate a class to be a data class. This decorator wraps the dataclasses.dataclass
    decorator, with the following changes to the generated __init__ method:

    • initialization method only processes keyword arguments
    • initialization method ignores unexpected keyword arguments
    • fields (with default values or not) can be declared in any order
    • optional fields default to None if no default value is specified
    """
    dc = dataclasses.dataclass(cls, init=False, **kwargs)
    if init:
        dc.__init__ = _datacls_init(dc)
    return dc


def make_datacls(
    cls_name: str,
    fields: Iterable[tuple[str, type] | tuple[str, type, dataclasses.Field]],
    init: bool = True,
    **kwargs,
) -> type:
    """
    Return a new dataclass. This function wraps the Python dataclasses.make_dataclass
    function, with the following changes to the generated __init__ method:

    • initialization method only processes keyword arguments
    • initialization method ignores unexpected keyword arguments
    • fields (with default values or not) can be declared in any order
    • Optional[...] fields default to None if no default value is specified

    Keyword arguments are passed on to the dataclasses.make_dataclass function.
    """
    dataclass = dataclasses.make_dataclass(
        cls_name=cls_name,
        fields=fields,
        init=False,
        **kwargs,
    )
    if init:
        dataclass.__init__ = _datacls_init(dataclass)
    return dataclass


def derive_datacls(
    cls_name: str,
    dataclass: type,
    *,
    include: set[str] = None,
    exclude: set[str] = None,
    append: Iterable[tuple[str, type] | tuple[str, type, dataclasses.Field]] = None,
    optional: set[str] | bool = False,
    **kwargs,
) -> type:
    """
    Return a new dataclass with fields derived from another dataclass.

    Parameters:
    • cls_name: the name of the new dataclass
    • dataclass: dataclass to derive fields from
    • include: the names of dataclass fields to include  [all]
    • exclude: the names of dataclass fields to exclude  [none]
    • optional: derived fields to make optional
    • append: iterable of field tuples to append to the derived dataclass

    If include is not specified, then all fields from the source dataclass are included. If
    exclude is not specified, then no fields from the source dataclass are excluded.

    The optional parameter can be a boolean value or an iterable of field names. If boolean,
    it specifies whether all derived fields should be made optional.

    The append parameter takes the same form as the fields parameter in the make_datacls
    function. These fields are not affected by the optional parameter.

    Any additional keyword arguments are passed on to the make_datacls function.
    """

    def _type(f):
        if optional is True or (optional is not False and f.name in optional):
            return f.type | None
        return f.type

    def _field(f):
        return dataclasses.field(
            default=f.default,
            default_factory=f.default_factory,
            init=f.init,
            repr=f.repr,
            hash=f.hash,
            compare=f.compare,
            metadata=f.metadata,
        )

    fields = [
        (f.name, _type(f), _field(f))
        for f in dataclasses.fields(dataclass)
        if (include is None or f.name in include) and (exclude is None or f.name not in exclude)
    ]

    for item in append or ():
        if len(item) == 2:
            item = (item[0], item[1], dataclasses.field())
        if len(item) != 3:
            raise TypeError(f"Invalid field: {item!r}")
        fields.append(item)

    return make_datacls(cls_name, fields=fields, **kwargs)


def derive_typeddict(
    type_name: str,
    source: Any,
    *,
    include: set[str] = None,
    exclude: set[str] = None,
    total: bool = True,
) -> type:
    """
    Generate a derived TypedDict from a source TypedDict or dataclass.

    Parameters:
    • type_name: the name of the new TypedDict type
    • source: TypedDict or dataclass to derive from
    • include: the names of keys to include  [all]
    • exclude: the names of keys to exclude  [none]
    • total: must all keys be present in the TypedDict
    """

    source, _ = split_annotations(source)

    if include is None:
        include = source.__annotations__.keys()

    if exclude is None:
        exclude = set()

    return TypedDict(
        type_name,
        {
            key: type
            for key, type in source.__annotations__.items()
            if key in include and key not in exclude
        },
        total=total,
    )


def copy_data(
    source: Any,
    target: Any,
    include: set[str] = None,
    exclude: set[str] = None,
) -> Any:
    """
    Creates a new instance of the target dataclass or TypedDict data type, populating it with
    data from the source dataclass or TypedDict instance.

    Parameters:
    • source: dataclass instance or mapping to copy data from
    • target: dataclass or TypedDict type or instance to copy data to
    • include: fields to include  [all]
    • exclude: fields to exclude  [none]

    If values to include are not specified, then all fields that source and target types have
    in common are copied. If fields to exclude are not specified, then no fields are excluded.

    This function makes no attempt to ensure that the types of data in source and target match.
    """

    if is_dataclass(source) and type(source) is not type:
        getter = lambda key: getattr(source, key, None)
    elif isinstance(source, Mapping):
        getter = lambda key: source.get(key, None)
    else:
        raise TypeError("source must be dataclass or mapping")

    if not (
        (is_dataclass(target) and type(target) is type)
        or (issubclass(target, dict) and hasattr(target, "__annotations__"))
    ):
        raise TypeError("target must be dataclass or TypedDict type")

    if include is None:
        include = source.__annotations__.keys() if is_dataclass(source) else source.keys()

    if exclude is None:
        exclude = set()

    keys = include & target.__annotations__.keys() - exclude
    kwargs = {key: getter(key) for key in keys}
    return target(**kwargs)


def redact_passwords(hint: Any, value: Any, redaction: str = "__REDACTED__") -> None:
    """
    Redact password fields in dataclass or TypedDict value.

    Parameters:
    • hint: type hint containing type of value to redact
    • value: value to be redacted
    • redaction: string replace redacted values with
    """
    if is_dataclass(value):
        getter, setter = functools.partial(getattr, value), functools.partial(setattr, value)
    elif isinstance(value, Mapping):
        getter, setter = value.get, value.__setitem__
    else:
        raise TypeError("type must be dataclass or TypedDict")
    value_type, _ = split_annotations(strip_optional(hint))
    for field_name, field_hint in value_type.__annotations__.items():
        field_type, field_annotations = split_annotations(strip_optional(field_hint))
        field_value = getter(field_name)
        if hasattr(field_type, "__annotations__") and (
            is_dataclass(field_value) or isinstance(field_value, Mapping)
        ):
            redact_passwords(field_hint, field_value)
        elif (
            field_value is not None
            and is_subclass(field_type, str)
            and Password in field_annotations
        ):
            setter(field_name, redaction)



"""Module for binary content streaming."""

from collections.abc import AsyncIterator


class Stream(AsyncIterator[bytes | bytearray]):
    """
    Base class to provide binary content through an asynchronous stream. The stream provides
    binary data through asynchronously iterable chunks of bytes or bytearray.

    During iteration, the stream determines the size of each chunk.

    Attributes:
    • content_type: the media (MIME) type of the stream
    • content_length: the length of the content, or None if unknown

    Much like a file, a stream is returned in an "open" state. The consumer must explicitly
    close it, either via by calling its `close` method, or using `async with`.
    """

    def __init__(self, content_type: str, content_length: int | None = None):
        self.content_type = content_type
        self.content_length = content_length

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc):
        await self.close()

    def __aiter__(self):
        return self

    async def __anext__(self) -> bytes | bytearray:
        raise NotImplementedError

    async def close(self) -> None:
        """
        Close the stream. Further attempts to iterate the stream will raise StopAsyncIteration.
        This method is idempotent; it is not an error to close a stream more than once.
        """
        raise NotImplementedError


class BytesStream(Stream):
    """
    Represents a bytes or bytearray object as an asynchronous byte stream. All content is
    returned in a single iteration.

    Parameters:
    • content: the data to be streamed
    • content_type: the MIME type of the data to be streamed
    """

    def __init__(
        self,
        content: bytes | bytearray,
        content_type: str = "application/octet-stream",
    ):
        super().__init__(content_type=content_type, content_length=len(content))
        self.content = content

    async def __anext__(self) -> bytes:
        if self.content is None:
            raise StopAsyncIteration
        result = self.content
        self.content = None
        return result

    async def close(self):
        self.content = None


async def read_stream(stream: Stream, limit: int | None = None) -> bytearray:
    """
    Read the entire content from a stream and return it in a byte array. This function does
    not close the stream after all bytes are read.

    Parameters:
    • stream: stream to read binary data from
    • limit: byte array length limit  [no limit]

    If the size byte array would exceed the specified limit, `ValueError` is raised.
    """

    result = bytearray()
    async for chunk in stream:
        result += chunk
        if limit and len(result) > limit:
            raise ValueError("byte array length limit exceeded")
    return result



"""Module that support validation of data."""

import asyncio
import dataclasses
import fondat.types
import inspect
import re
import types
import typing
import wrapt

from collections.abc import Callable, Iterable, Mapping
from contextlib import contextmanager
from fondat.types import is_instance, is_subclass, split_annotations
from types import NoneType
from typing import Any, TypeVar


class ValidationError(ValueError):
    """Error raised when validation fails."""

    __slots__ = {"message", "path"}

    def __init__(self, message: str | None = None, path: list[str | int] | None = None):
        self.message = message
        self.path = path

    def __repr__(self):
        return f"{self.__class__.__name__}({self.message!r}, {self.path!r})"

    def __str__(self):
        result = []
        if self.message is not None:
            result.append(str(self.message))
        if self.path:
            result.append(f"({'.'.join((str(a) for a in self.path))})")
        return " ".join(result)

    @staticmethod
    @contextmanager
    def path_on_error(path: list[str | int] | str | int):
        """Context manager to specify error path in the event that a DecodeError is raised."""
        try:
            yield
        except ValidationError as ve:
            if ve.path is None:
                ve.path = []
            match path:
                case str() | int():
                    ve.path.insert(0, path)
                case list():
                    ve.path = path + ve.path
            raise


class Validator:
    """Base class for type annotation that performs validation."""

    def validate(self, value: Any) -> None:
        raise NotImplementedError


class MinLen(Validator):
    """Type annotation that validates a value has a minimum length."""

    __slots__ = {"value"}

    def __init__(self, value: int):
        self.value = value

    def validate(self, value: Any) -> None:
        if len(value) < self.value:
            raise ValidationError(f"mininum length: {self.value}")

    def __repr__(self):
        return f"MinLen({self.value})"


class MaxLen(Validator):
    """Type annotation that validates a value has a maximum length."""

    __slots__ = {"value"}

    def __init__(self, value: int):
        self.value = value

    def validate(self, value: Any) -> None:
        if len(value) > self.value:
            raise ValidationError(f"maximum length: {self.value}")

    def __repr__(self):
        return f"MaxLen({self.value})"


class MinValue(Validator):
    """Type annotation that validates a value has a minimum value."""

    __slots__ = {"value"}

    def __init__(self, value: Any):
        self.value = value

    def validate(self, value: Any) -> None:
        if value < self.value:
            raise ValidationError(f"minimum value: {self.value}")

    def __repr__(self):
        return f"MinValue({self.value})"


class MaxValue(Validator):
    """Type annotation that validates a value has a maximum value."""

    __slots__ = {"value"}

    def __init__(self, value: Any):
        self.value = value

    def validate(self, value: Any) -> None:
        if value > self.value:
            raise ValidationError(f"maximum value: {self.value}")

    def __repr__(self):
        return f"MaxValue({self.value})"


class Pattern(Validator):
    """
    Type annotation that validates a value matches a pattern.

    Parameters:
    • pattern: pattern object or string to match
    """

    __slots__ = {"pattern"}

    def __init__(self, pattern: str | re.Pattern):
        self.pattern = re.compile(pattern) if isinstance(pattern, str) else pattern

    def validate(self, value: Any) -> None:
        if not self.pattern.match(value):
            raise ValidationError(f"pattern: {self.pattern.pattern}")

    def __repr__(self):
        return f"Pattern({self.pattern})"


def _validate_union(value, args):
    if value is None and NoneType in args:
        return
    for arg in args:
        try:
            return validate(value, arg)
        except ValidationError as e:
            continue
    raise ValidationError(f"expecting: union of {args}; received: {type(value)} ({value})")


def _validate_literal(value, args):
    for arg in args:
        if arg == value and type(arg) is type(value):
            return
    raise ValidationError(f"expecting one of: {args}; received: {value}")


@contextmanager
def validation_error_path(segment: Any):
    try:
        yield
    except ValidationError as ve:
        ve.path = (ve.path or []).insert(0, segment)
        raise


def _validate_typeddict(value, python_type):
    for item_key, item_type in typing.get_type_hints(python_type, include_extras=True).items():
        try:
            with validation_error_path(item_key):
                validate(value[item_key], item_type)
        except KeyError:
            if item_key in python_type.__required_keys__:
                raise ValidationError("required", path=[item_key])


def _validate_mapping(value, python_type, args):
    key_type, value_type = args
    for key, value in value.items():
        validate(key, key_type)
        with validation_error_path(key):
            validate(value, value_type)


def _validate_tuple(value, python_type, args):
    if len(args) == 2 and args[1] is Ellipsis:
        item_type = args[0]
        index = 0
        for n in range(len(value)):
            with validation_error_path(n):
                validate(value[n], item_type)
        index += 1
    elif len(value) != len(args):
        raise ValidationError(
            f"expecting tuple[{', '.join(str(arg) for arg in args)}]; received: {value}"
        )
    else:
        for n in range(len(args)):
            with validation_error_path(n):
                validate(value[n], args[n])


def _validate_iterable(value, python_type, args):
    item_type = args[0]
    index = 0
    for item_value in value:
        with validation_error_path(index):
            validate(item_value, item_type)
        index += 1


def _validate_typevar(value, python_type):
    validate(value, fondat.types.resolve_typevar(python_type))


def _validate_dataclass(value, python_type, origin):
    dc_type = python_type if not origin else origin
    with fondat.types.capture_typevars(python_type):
        for attr_name, attr_type in typing.get_type_hints(dc_type, include_extras=True).items():
            with validation_error_path(attr_name):
                validate(getattr(value, attr_name), attr_type)


def validate(value: Any, type_hint: Any) -> NoneType:
    """Validate a value."""

    python_type, annotations = split_annotations(type_hint)
    origin = typing.get_origin(python_type)
    args = typing.get_args(python_type)

    # validate using specified validator annotations
    for annotation in annotations:
        if isinstance(annotation, Validator):
            annotation.validate(value)

    if python_type is Any:
        return
    elif isinstance(python_type, TypeVar):
        return _validate_typevar(value, python_type)

    match origin:
        case types.UnionType | typing.Union:
            return _validate_union(value, args)
        case typing.Literal:
            return _validate_literal(value, args)

    # TypedDict
    if is_subclass(python_type, dict) and hasattr(python_type, "__annotations__"):
        origin = dict

    # basic type validation
    if origin and not is_instance(value, origin):
        raise ValidationError(f"expecting {origin.__name__}; received {type(value)}")
    elif not origin and not is_instance(value, python_type):
        raise ValidationError(f"expecting {python_type}; received {type(value)}")
    elif python_type is int and is_instance(value, bool):  # bool is subclass of int
        raise ValidationError("expecting int; received bool")
    elif is_subclass(origin, Iterable) and is_instance(value, (str, bytes, bytearray)):
        raise ValidationError(f"expecting Iterable; received {type(value)}")

    # structured type validation
    if is_subclass(python_type, dict) and hasattr(python_type, "__annotations__"):
        return _validate_typeddict(value, python_type)
    elif is_subclass(origin, Mapping):
        return _validate_mapping(value, python_type, args)
    elif is_subclass(origin, tuple):
        return _validate_tuple(value, python_type, args)
    elif is_subclass(origin, Iterable):
        return _validate_iterable(value, python_type, args)
    elif dataclasses.is_dataclass(python_type) or dataclasses.is_dataclass(origin):
        return _validate_dataclass(value, python_type, origin)


def validate_arguments(callable: Callable):
    """Decorate a function or coroutine to validate its arguments using type annotations."""

    sig = inspect.signature(callable)

    positional_params = [
        p.name
        for p in sig.parameters.values()
        if p.kind in {p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD}
    ]

    def _validate(instance, args, kwargs):
        hints = typing.get_type_hints(callable, include_extras=True)
        if instance:
            args = (instance, *args)
        params = {
            **{p: v for p, v in zip(positional_params, args)},
            **kwargs,
        }
        for param in (p for p in sig.parameters.values() if p.name in params):
            if hint := hints.get(param.name):
                with validation_error_path(param.name):
                    validate(params[param.name], hint)

    if asyncio.iscoroutinefunction(callable):

        @wrapt.decorator
        async def decorator(wrapped, instance, args, kwargs):
            _validate(instance, args, kwargs)
            return await wrapped(*args, **kwargs)

    else:

        @wrapt.decorator
        def decorator(wrapped, instance, args, kwargs):
            _validate(instance, args, kwargs)
            return wrapped(*args, **kwargs)

    return decorator(callable)


def validate_return_value(callable: Callable):
    """Decorate a function or coroutine to validate its return value using type hints."""

    return_type = typing.get_type_hints(callable, include_extras=True).get("return")

    def _validate(result):
        if return_type is not None:
            with validation_error_path("return"):
                validate(result, return_type)

    if asyncio.iscoroutinefunction(callable):

        @wrapt.decorator
        async def decorator(wrapped, instance, args, kwargs):
            result = await wrapped(*args, **kwargs)
            _validate(result)
            return result

    else:

        @wrapt.decorator
        def decorator(wrapped, instance, args, kwargs):
            result = wrapped(*args, **kwargs)
            _validate(result)
            return result

    return decorator(callable)


def is_valid(value: Any, type_hint: Any) -> bool:
    """Return if a value is valid for specified type."""

    try:
        validate(value, type_hint)
    except ValidationError:
        return False
    return True



"""
Module to support pagination of items.

For an operation that returns a large set of items, it can be expensive to return all items in
a single response. In this case, the operation should return items in pages, requiring multiple
calls to the operation to retrieve all items.

Paginated items are provided through a page dataclass, which minimally contains:

  • an iterable object of `items` in the page
  • an opaque `cursor` value to retrieve the next page

The caller will initially pass no cursor value to an operation, resulting in the first page
of items. When generating the page, if there are additional items to be returned, the page
cursor should contain an opaque value that can be used to request the subsequent page. The
last page should contain no cursor value, indicating there are no further items to request.

An operation should establish a reasonable limit of items it would return in each page. If
appropriate, the operation can also expose a `limit` parameter, which allows a caller to
request a number of items to be returned in the page. Regardless of how many items are
requested, the operation should be free to decide how many items to actually return.
"""

from collections.abc import Callable, Coroutine, Iterable
from dataclasses import dataclass
from typing import Any, Generic, TypeVar


class PaginationError(Exception):
    """
    Error raised if pagination could not be performed.

    An example of a pagination error is a loss of indexing: pages have drifted due to changes
    to the underlying data between page requests, and the continuation point of the next page
    cannot be reconstructed.
    """


# type aliases
Item = TypeVar("Item")
Cursor = bytes | None


@dataclass
class Page(Generic[Item]):
    """A paginated result."""

    items: Iterable[Item]
    cursor: Cursor = None


async def paginate(operation: Callable[..., Coroutine[Any, Any, Any]], /, **kwargs):
    """
    Wraps a paginated resource operation with an asynchronous generator that iterates through
    all items. The wrapped resource operation must return a page dataclass and accept a
    `cursor` parameter.

    Parameters:
    • operation: resource operation to wrap with generator
    • kwargs: keyword arguments to pass to resource operation
    """
    cursor = {}
    while cursor is not None:
        page = await operation(**kwargs, **cursor)
        cursor = {"cursor": page.cursor} if page.cursor is not None else None
        for item in page.items:
            yield item



"""Module to manage data in a SQLite database."""

import aiosqlite
import asyncio
import contextvars
import fondat.codec
import fondat.sql
import logging
import sqlite3
import types
import typing
import uuid

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from fondat.codec import Codec, DecodeError, EncodeError
from fondat.sql import Expression, Param
from fondat.types import is_optional, is_subclass, literal_values, strip_annotations
from types import NoneType
from typing import Any, Literal, TypeVar


_logger = logging.getLogger(__name__)


PT = TypeVar("PT")  # Python type hint
ST = TypeVar("ST")  # SQL type hint


class SQLiteCodec(Codec[PT, ST]):
    """Base class for SQLite codecs."""

    _cache = {}


class BlobCodec(SQLiteCodec[bytes | bytearray, bytes]):
    """
    Codec that encodes/decodes a value to/from a SQL BLOB. Supports the following types:
    bytes, bytearray.
    """

    sql_type = "BLOB"

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, bytes | bytearray)

    def encode(self, value: bytes | bytearray) -> bytes:
        if not isinstance(value, bytes | bytearray):
            raise EncodeError
        return bytes(value)

    def decode(self, value: bytes) -> bytes | bytearray:
        if not isinstance(value, bytes):
            raise DecodeError
        return self.python_type(value)


class IntegerCodec(SQLiteCodec[int | bool, int]):
    """
    Codec that encodes/decodes a value to/from a SQL INTEGER. Supports the following types:
    int, bool.
    """

    sql_type = "INTEGER"

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, int | bool)

    def encode(self, value: int | bool) -> int:
        if not isinstance(value, int | bool):
            raise EncodeError
        return int(value)

    def decode(self, value: int) -> int | bool:
        if not isinstance(value, int):
            raise DecodeError
        return self.python_type(value)


class RealCodec(SQLiteCodec[float, float]):
    """
    Codec that encodes/decodes a value to/from a SQL REAL. Supports the following type: float.
    """

    sql_type = "REAL"

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, float)

    def encode(self, value: float) -> float:
        if not isinstance(value, float):
            raise EncodeError
        return value

    def decode(self, value: float) -> float:
        if not isinstance(value, float):
            raise DecodeError
        return value


class UnionCodec(SQLiteCodec[PT, Any]):
    """
    Codec that encodes/decodes a UnionType, Union or optional value to/from a compatible SQL
    value. For an optional type, it will use the codec for its type, otherwise it will
    encode/decode as TEXT.
    """

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return typing.get_origin(python_type) in {typing.Union, types.UnionType}

    def __init__(self, python_type: type[PT]):
        super().__init__(python_type)
        raw_type = strip_annotations(python_type)
        args = typing.get_args(raw_type)
        self.is_nullable = is_optional(raw_type)
        args = [a for a in args if a is not NoneType]
        self.codec = SQLiteCodec.get(args[0]) if len(args) == 1 else TextCodec(python_type)
        self.sql_type = self.codec.sql_type

    def encode(self, value: PT) -> Any:
        if value is None:
            return None
        return self.codec.encode(value)

    def decode(self, value: Any) -> PT:
        if value is None and self.is_nullable:
            return None
        return self.codec.decode(value)


class LiteralCodec(SQLiteCodec[PT, Any]):
    """
    Codec that encodes/decodes a Literal value to/from a compatible SQL value. If all literal
    values share the same type, then it will use a codec for that type, otherwise it will
    encode/decode as TEXT.
    """

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return typing.get_origin(python_type) is Literal

    def __init__(self, python_type: type[PT]):
        super().__init__(python_type)
        self.literals = literal_values(python_type)
        types = list({type(literal) for literal in self.literals})
        self.codec = SQLiteCodec.get(types[0]) if len(types) == 1 else TextCodec(python_type)
        self.is_nullable = is_optional(python_type) or None in self.literals
        self.sql_type = self.codec.sql_type

    def encode(self, value: PT) -> Any:
        if value is None:
            return None
        return self.codec.encode(value)

    def decode(self, value: Any) -> PT:
        if value is None and self.is_nullable:
            return None
        result = self.codec.decode(value)
        if result not in self.literals:
            raise DecodeError
        return result


class TextCodec(SQLiteCodec[PT, Any]):
    """
    Codec that encodes/decodes a value to/from a SQL TEXT. This is the "fallback" codec,
    which handles any type not handled by any other codec.
    """

    sql_type = "TEXT"

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        for other in (c for c in SQLiteCodec.__subclasses__() if c is not TextCodec):
            if other.handles(python_type):
                return False
        return True

    def __init__(self, python_type):
        super().__init__(python_type)
        self.string_codec = fondat.codec.StringCodec.get(python_type)

    def encode(self, value: PT) -> str:
        return self.string_codec.encode(value)

    def decode(self, value: str) -> PT:
        return self.string_codec.decode(value)


class _Results(AsyncIterator[Any]):

    __slots__ = {"statement", "result", "rows", "codecs"}

    def __init__(self, statement, result, rows):
        self.statement = statement
        self.result = result
        self.rows = rows
        self.codecs = {
            k: SQLiteCodec.get(t)
            for k, t in typing.get_type_hints(result, include_extras=True).items()
        }

    def __aiter__(self):
        return self

    async def __anext__(self):
        row = await self.rows.__anext__()
        return self.result(**{k: self.codecs[k].decode(row[k]) for k in self.codecs})


@asynccontextmanager
async def _async_null_context():
    yield


class Database(fondat.sql.Database):
    """
    Manages access to a SQLite database.

    Parameter:
    • path: path to SQLite database file
    """

    __slots__ = {"path", "_conn", "_txn"}

    def __init__(self, path: str):
        super().__init__()
        self.path = path
        self._conn = contextvars.ContextVar("fondat_sqlite_conn", default=None)
        self._txn = contextvars.ContextVar("fondat_sqlite_txn", default=None)
        self._task = contextvars.ContextVar("fondat_sqlite_task", default=None)

    @asynccontextmanager
    async def connection(self):
        task = asyncio.current_task()
        if self._conn.get() and self._task.get() is task:
            yield  # connection already established
            return
        _logger.debug("open connection")
        self._task.set(task)
        connection = await aiosqlite.connect(self.path)
        connection.row_factory = sqlite3.Row
        self._conn.set(connection)
        try:
            yield
        finally:
            _logger.debug("close connection")
            self._conn.set(None)
            try:
                await connection.close()
            except Exception as e:
                _logger.exception("error closing connection")

    @asynccontextmanager
    async def transaction(self):
        txid = f"_{uuid.uuid4().hex}"
        _logger.debug("transaction begin %s", txid)
        token = self._txn.set(txid)

        async def commit():
            _logger.debug("transaction commit %s", txid)
            await connection.execute(f"RELEASE SAVEPOINT {txid};")

        async def rollback():
            _logger.debug("transaction rollback %s", txid)
            await connection.execute(f"ROLLBACK TO SAVEPOINT {txid};")

        async with self.connection():
            connection = self._conn.get()
            await connection.execute(f"SAVEPOINT {txid};")
            try:
                yield
            except GeneratorExit:  # explicit cleanup of asynchronous generator
                await commit()
            except Exception:
                await rollback()
                raise
            else:
                await commit()
            finally:
                self._txn.reset(token)

    async def execute(
        self,
        statement: Expression,
        result: type = None,
    ) -> AsyncIterator[Any] | None:
        if not self._txn.get():
            raise RuntimeError("transaction context required to execute statement")
        text = []
        args = []
        for fragment in statement:
            match fragment:
                case str():
                    text.append(fragment)
                case Param():
                    text.append("?")
                    args.append(SQLiteCodec.get(fragment.type).encode(fragment.value))
                case _:
                    raise ValueError(f"unexpected fragment: {fragment}")
        results = await self._conn.get().execute("".join(text), args)
        if result is not None:  # expecting a result
            return _Results(statement, result, results.__aiter__())

    def sql_type(self, type: Any) -> str:
        return SQLiteCodec.get(type).sql_type



"""Comma separated values encoding module."""

import dataclasses

from collections.abc import Iterable, Mapping
from fondat.codec import DecodeError, StringCodec
from fondat.data import derive_typeddict
from fondat.types import is_optional, strip_annotations
from typing import Any, Generic, TypeVar, get_type_hints, is_typeddict


def _round(value: Any, precision: int | None) -> str:
    if precision is None:  # floating point
        value = str(value)
        if "." in value:
            value = value.rstrip("0").rstrip(".")
        return value
    if precision == 0:
        return str(round(value))
    return f"{{:.{precision}f}}".format(value)


T = TypeVar("T")


class CurrencyCodec(Generic[T]):
    """
    String codec that encodes/decodes a number as a currency value; optionally encodes with
    fixed-point precision.

    Parameters:
    • python_type: type of the value to be encoded/decoded
    • prefix: currency prefix (e.g. "$")
    • suffix: currency suffix
    • precision: round encoded value to number of digits  [floating point]
    """

    def __init__(
        self,
        python_type: Any,
        prefix: str = "",
        suffix: str = "",
        precision: int | None = None,
    ):
        self.prefix = prefix
        self.suffix = suffix
        self.precision = precision
        self.codec = StringCodec.get(python_type)

    def encode(self, value: T) -> str:
        return (
            f"{self.prefix}{_round(value, self.precision)}{self.suffix}"
            if value is not None
            else ""
        )

    def decode(self, value: str) -> T:
        result = self.codec.decode(value.lstrip(self.prefix).rstrip(self.suffix))
        if self.precision is not None:
            result = round(result, self.precision)
        return result


class PercentCodec(Generic[T]):
    """
    String codec that encodes/decodes a fractional value as a percentage string with
    fixed-point precision.

    Parameters:
    • python_type: type of the value to be encoded/decoded
    • precision: round encoded value to number of digits
    """

    def __init__(self, python_type: Any, precision: int):
        self.precision = precision
        self.codec = StringCodec.get(python_type)

    def encode(self, value: T) -> str:
        return f"{_round(value * 100, self.precision)}%"

    def decode(self, value: str) -> T:
        result = self.codec.decode(value.rstrip("%")) / 100
        if self.precision is not None:
            result = round(result, self.precision + 2)
        return result


class FixedCodec(Generic[T]):
    """
    String codec encodes/decodes a number with fixed-point precision.

    Parameter:
    • python_type: type of the value to be encoded/decoded
    • precision: round encoded value to number of digits
    """

    def __init__(self, python_type: Any, precision: int):
        self.precision = precision
        self.codec = StringCodec.get(python_type)

    def encode(self, value: T) -> str:
        return _round(value, self.precision) if value is not None else ""

    def decode(self, value: str) -> T:
        result = self.codec.decode(value)
        if self.precision is not None:
            result = round(result, self.precision)
        return result


class TypedDictCodec(Generic[T]):
    """
    Codec that encodes/decodes a typed dictionary to/from a CSV row. A CSV row is a list of
    strings.

    Parameters:
    • typeddict: TypedDict type to encode/decode
    • columns: ordered column names
    • keys: mapping between columns and dictionary keys
    • codecs: mapping between columns and codecs

    Attribute:
    • columns: ordered column names

    The columns parameter and attribute specifies the names of CSV columns, and the order they
    are encoded in a row. If the columns parameter is omitted, then columns will be all
    TypedDict keys, in the order they are defined in the TypedDict.

    The keys mapping specifies the mapping between columns and dictionary keys. If no key
    for a given column is specified, then the column will map the to dictionary key of the
    same name.

    The codecs mapping specifies which codecs are used to encode columns. If no codec for a
    given column is specified, then the default string codec for its associated type is used.
    """

    def __init__(
        self,
        typeddict: Any,
        columns: Iterable[str] | None = None,
        keys: Mapping[str, str] | None = None,
        codecs: Mapping[str, Any] | None = None,
    ):
        typeddict = strip_annotations(typeddict)
        if not is_typeddict(typeddict):
            raise TypeError("typeddict parameter must be a TypedDict")

        hints = get_type_hints(typeddict, include_extras=True)

        self.columns = columns or tuple(k for k in hints.keys())

        if keys is None:
            keys = {column: column for column in self.columns}

        self._keys = {column: key for column, key in keys.items() if key in hints}

        if codecs is None:
            codecs = {}

        self._codecs = {
            column: codecs.get(column, StringCodec.get(hints[self._keys[column]]))
            for column in self.columns
            if column in self._keys
        }

        self._optional = {key for key, hint in hints.items() if is_optional(hint)}

    def encode(self, value: T) -> list[str]:
        """
        Encode from TypedDict value to CSV row. If a field value is None, it will be
        represented in a column as an empty string.
        """
        return [
            self._codecs[c].encode(value.get(self._keys[c], None)) if c in self._codecs else ""
            for c in self.columns
        ]

    def decode(self, values: list[str]) -> T:
        """
        Decode from CSV row to TypedDict value. If a column to decode contains an empty
        string value, it will be represented as None if the associated TypedDict value is
        optional.
        """
        result = {}
        for n in range(len(self.columns)):
            column = self.columns[n]
            key = self._keys.get(column, None)
            if key is not None:
                try:
                    value = values[n]
                except IndexError:
                    value = ""
                if value == "" and key in self._optional:
                    result[key] = None
                else:
                    with DecodeError.path_on_error(column):
                        result[key] = self._codecs[column].decode(value)
        return result


class DataclassCodec(Generic[T]):
    """
    Codec that encodes/decodes a dataclass to/from a CSV row. A CSV row is a list of strings.

    Parameters:
    • dataclass: dataclass type to encode/decode
    • columns: ordered column names
    • fields: mapping between row columns and dataclass fields
    • codecs: mapping between columns and codecs

    Attribute:
    • columns: ordered column names

    The columns parameter specifies the names of CSV columns, and the order they are encoded
    in a row. If the columns parameter is omitted, then columns will be all dataclass
    fields, in the order they are defined in the dataclass.

    The fields mapping specifies the mapping between column names and dictionary keys. If no
    mapping for a given column is specified, then the column will map to the field name of
    the same name.

    The codecs mapping specifies which codecs are used to encode columns. If no mapping for a
    given column is provided, then the default codec for its associated field is used.
    """

    def __init__(
        self,
        dataclass: Any,
        columns: Iterable[str] = None,
        fields: Mapping[str, str] = None,
        codecs: Mapping[str, Any] = None,
    ):
        dataclass = strip_annotations(dataclass)
        if not dataclasses.is_dataclass(dataclass):
            raise TypeError("dataclass parameter must be a dataclass")

        self.dataclass = dataclass
        self.codec = TypedDictCodec(
            typeddict=derive_typeddict("TD", dataclass),
            columns=columns,
            keys=fields,
            codecs=codecs,
        )

    @property
    def columns(self) -> Iterable[str]:
        return self.codec.columns

    def encode(self, value: T) -> list[str]:
        """
        Encode from dataclass value to CSV row. If a field value is None, it will be
        represented in a column as an empty string.
        """
        return self.codec.encode(dataclasses.asdict(value))

    def decode(self, values: list[str]) -> T:
        """
        Decode from CSV row to dataclass value. If a column to decode contains an empty
        string value, it will be represented as None if the associated field is optional.
        """
        return self.dataclass(**self.codec.decode(values))



"""Fondat string module."""

import re

from collections.abc import Callable, Coroutine
from typing import Any


class Template:
    """
    A string class for asynchronous substitution of template expressions.

    Parameters:
    • template: template string containing expressions to be substituted

    A substition expression in the template takes the form "${...}". For example, if a
    template contains "${foo}" substitution expression, then "foo" is the value that is passed
    to a resolver coroutine to be resolved. A subsitution expression cannot span lines
    within the template string.
    """

    _pattern = re.compile(r"\$\{(.*?)}")

    def __init__(self, template: str):
        self.template = template

    async def _resolve(
        self, template: str, resolver: Callable[[str], Coroutine[Any, Any, str | None]]
    ) -> str:
        segments = []
        for line in template.splitlines(keepends=True):
            index = 0
            while match := self._pattern.search(line[index:]):
                group = match.group(1)
                resolved = None
                if resolved := await resolver(group):
                    span = match.span()
                    segments.append(line[index : span[0]])
                    segments.append(await self._resolve(resolved, resolver))
                    index = span[1]
                    break
                if not resolved:
                    raise ValueError(f"could not resolve ${{{group}}} in template")
            segments.append(line[index:])
        return "".join(segments)

    async def resolve(self, resolver: Callable[[str], Coroutine[Any, Any, str | None]]):
        """
        Return a new string with resolved template substitutions.

        Parameters:
        • resolver: coroutine function to resolve substitution expressions

        A resolver coroutine function takes a string parameter and returns a string. If the
        resolver cannot resolve an expression, it should return None or raise a ValueError with
        a descriptive error message.

        It is valid for a resolver to return a value that itself contains one or more
        substitition expressions; they will then be resolved.
        """
        return await self._resolve(self.template, resolver)



"""Document partial modification (patch) module."""

import collections.abc

from copy import deepcopy
from fondat.codec import JSONCodec
from typing import Any


def _json_merge_patch(target, patch):
    if isinstance(patch, collections.abc.Mapping):
        if not isinstance(target, collections.abc.Mapping):
            target = {}
        for key, value in patch.items():
            if value is None:
                target.pop(key, None)
            else:  # recursive
                target[key] = _json_merge_patch(target.get(key), value)
        return target
    else:
        return deepcopy(patch)


def json_merge_patch(*, value: Any, type: type = Any, patch: Any) -> Any:
    """
    Return the result of applying a JSON Merge Patch document to the JSON representation of
    a specified value, per RFC 7386.

    Parameters:
    • value: value to be patched
    • type: type of value to be patched
    • patch: JSON Merge Patch document to apply to value
    """
    codec = JSONCodec.get(type)
    result = codec.decode(_json_merge_patch(codec.encode(value), patch))
    return result


def _json_merge_diff(old: Any, new: Any) -> Any:
    if isinstance(old, collections.abc.Mapping) and isinstance(new, collections.abc.Mapping):
        diff = {}
        for key in new:
            if key in old:
                old_value = old[key]
                new_value = new[key]
                if new_value != old_value:  # recursive
                    if (d := _json_merge_diff(old_value, new_value)) != {}:
                        diff[key] = d
            else:
                diff[key] = new[key]
        for key in old:
            if key not in new:
                diff[key] = None
        return diff
    else:
        return new


def json_merge_diff(*, old: Any, new: Any, type: Any = Any) -> Any:
    """
    Return a JSON Merge Patch document per RFC 7386, the result of comparing the JSON
    representations of specified old and new values.
    """
    codec = JSONCodec.get(type)
    return _json_merge_diff(codec.encode(old), codec.encode(new))



"""File I/O module."""

import builtins
import logging
import mimetypes

from contextlib import suppress
from fondat.codec import BinaryCodec, DecodeError, StringCodec
from fondat.error import ConflictError, InternalServerError, NotFoundError
from fondat.http import AsBody
from fondat.resource import operation, resource
from fondat.stream import Stream, read_stream
from pathlib import Path
from typing import Annotated, BinaryIO, Generic, TypeVar
from urllib.parse import quote, unquote


_logger = logging.getLogger(__name__)


K = TypeVar("K")
V = TypeVar("V")


_content_encoding_map = {
    "br": "application/x-br",
    "bzip2": "application/x-bzip2",
    "compress": "application/x-compress",
    "gzip": "application/gzip",
    "xz": "application/x-xz",
}


def _content_type(url: str) -> str:
    content_type, content_encoding = mimetypes.guess_type(url)
    if content_encoding:
        return _content_encoding_map.get(content_encoding, content_encoding)
    if content_type:
        return content_type
    return "application/octet-stream"


class FileStream(Stream):
    """
    Represents a file as an asynchronous byte stream.

    Parameters:
    • path: ...
    • content_type: ...

    If content_type is not specified, this class will attempt to guess the content type
    from the filename extension.
    """

    def __init__(self, path: Path, content_type: str | None = None):
        self.file = path.open("rb")
        self.file.seek(0, 2)
        if content_type is None:
            content_type = _content_type(path.name)
        content_length = self.file.tell()
        self.file.seek(0, 0)
        super().__init__(content_type, content_length)

    async def __anext__(self) -> bytes:
        if self.file:
            chunk = self.file.read1(1048576)  # 1 MiB
            if len(chunk):
                return chunk
            await self.close()
        raise StopAsyncIteration

    async def close(self):
        with suppress(Exception):
            self.file.close()
        self.file = None


@resource
class DirectoryResource(Generic[K, V]):
    """
    Resource that represents files in a directory.

    Parameters:
    • path: path to directory where files are stored
    • key_type: type of key to identify each file
    • value_type: type of value stored in each file
    • extenson: filename extension to append (including dot)
    """

    def __init__(
        self,
        path: Path,
        key_type: type[K] = str,
        value_type: type[V] = Stream,
        extension: str | None = None,
    ):
        self.path = path.expanduser()
        if not self.path.is_dir():
            raise FileNotFoundError(f"directory not found: {self.path}")
        if not getattr(key_type, "__hash__", None):
            raise TypeError("invalid key_type: {key_type}")
        self.key_type = key_type
        self.value_type = value_type
        self.extension = extension
        self.key_codec = StringCodec.get(key_type)

    @operation
    async def get(self) -> set[K]:
        """Return list of file keys."""
        try:
            keys = set()
            for name in (
                path.name[: -len(self.extension)] if self.extension else path.name
                for path in self.path.iterdir()
                if path.is_file()
                and not path.name.endswith(".__tmp__")
                and (not self.extension or path.name.endswith(self.extension))
            ):
                try:
                    keys.add(self.key_codec.decode(unquote(name)))
                except DecodeError:
                    pass  # ignore incompatible file names
            return keys
        except Exception as e:
            raise InternalServerError from e

    def __getitem__(self, key: K) -> "FileResource[V]":
        return FileResource(
            self.path.joinpath(
                quote(self.key_codec.encode(key), safe="") + (self.extension or "")
            ),
            self.value_type,
        )


@resource
class FileResource(Generic[V]):
    """
    Resource that represents a file.

    Parameters:
    • path: location of file
    • type: type of value stored in file
    """

    def __init__(
        self,
        path: Path,
        type: builtins.type[V] = Stream,
    ):
        self.path = path.expanduser()
        self.type = type
        self.codec = BinaryCodec.get(type) if type is not Stream else None

    @operation
    async def get(self) -> V:
        """Read resource."""
        try:
            stream = FileStream(self.path)
            if self.type is Stream:
                return stream
            async with stream:  # close stream after reading
                return self.codec.decode(await read_stream(stream))
        except FileNotFoundError as fnfe:
            raise NotFoundError from fnfe
        except Exception as e:
            raise InternalServerError from e

    @operation
    async def put(self, value: Annotated[V, AsBody]):
        """Write resource."""
        try:
            tmp = self.path.with_name(f"{self.path.name}.__tmp__")
            with tmp.open("xb") as file:
                if self.type is Stream:
                    async with value as stream:  # close stream after reading
                        await write_stream(stream, file)
                else:
                    file.write(self.codec.encode(value))
            tmp.replace(self.path)
        except FileExistsError as e:
            raise ConflictError from e
        except Exception as e:
            raise InternalServerError from e

    @operation
    async def delete(self):
        """Delete resource."""
        try:
            self.path.unlink()
        except FileNotFoundError as fnfe:
            raise NotFoundError from fnfe
        except Exception as e:
            raise InternalServerError from e


async def write_stream(stream: Stream, file: BinaryIO):
    """Write a stream to a binary file-like object."""
    async for chunk in stream:
        file.write(chunk)



"""Module to expose resources through HTTP."""

#  TODO: In docstring, add description of routing through resource(s) to an operation.

import asyncio
import base64
import fondat.error
import fondat.resource
import fondat.types
import functools
import http
import http.cookies
import inspect
import logging
import multidict
import typing

from collections import namedtuple
from collections.abc import Callable, Iterable, MutableSequence
from fondat.codec import BinaryCodec, DecodeError, StringCodec
from fondat.error import (
    BadRequestError,
    InternalServerError,
    MethodNotAllowedError,
    NotFoundError,
)
from fondat.security import Scheme
from fondat.stream import BytesStream, Stream, read_stream
from fondat.types import is_optional, is_subclass
from typing import Annotated, Any, TypedDict


_logger = logging.getLogger(__name__)


Headers = multidict.CIMultiDict
Cookies = http.cookies.SimpleCookie
Query = multidict.MultiDict


class Message:
    """
    Base class for HTTP request and response.

    Parameters and attributes:
    • headers: multi-value dictionary to store headers; excludes cookies
    • cookies: dictionary containing message cookies
    • body: stream message body, or None if no body
    """

    def __init__(
        self,
        *,
        headers: Headers | None = None,
        cookies: Cookies | None = None,
        body: Stream | None = None,
    ):
        super().__init__()
        self.headers = headers or Headers()
        self.cookies = cookies or Cookies()
        self.body = body

    def __repr__(self):
        return f"Message(headers={self.headers}, cookies={self.cookies}, body={self.body})"


class Request(Message):
    """
    HTTP request.

    Parameters and attributes:
    • headers: multi-value dictionary to store headers; excludes cookies
    • cookies: simple cookie object to store request cookies
    • body: stream for request body, or None
    • method: the HTTP method name, in upper case
    • path: HTTP request target excluding query string
    • version: version of the incoming HTTP request
    • query: multi-value dictionary to store query string parameters
    """

    def __init__(
        self,
        *,
        headers: Headers | None = None,
        cookies: Cookies | None = None,
        body: Stream | None = None,
        method: str = "GET",
        path: str = "/",
        version: str = "1.1",
        query: Query | None = None,
    ):
        super().__init__(headers=headers, cookies=cookies, body=body)
        self.method = method
        self.path = path
        self.version = version
        self.query = query or Query()

    def __repr__(self):
        return (
            f"Request(headers={self.headers}, cookies={self.cookies}, body={self.body}, "
            f"method={self.method}, path={self.path}, version={self.version}, "
            f"query={self.query})"
        )


class Response(Message):
    """
    HTTP response.

    Parameters and attributes:
    • headers: multi-value dictionary to store headers; excludes cookies
    • cookies: dictionary containing response cookies
    • body: stream for response body, or None
    • status: HTTP status code
    """

    def __init__(
        self,
        *,
        headers: Headers | None = None,
        cookies: Cookies | None = None,
        body: Stream | None = None,
        status: int = http.HTTPStatus.OK.value,
    ):
        super().__init__(headers=headers, cookies=cookies, body=body)
        self.status = status

    def __repr__(self):
        return (
            f"Response(headers={self.headers}, cookies={self.cookies}, body={self.body}, "
            f"status={self.status})"
        )


class Chain:
    """
    A chain of zero or more filters, terminated by a single handler.

    A handler is a coroutine function that receives a request and returns a response. A chain
    is itself a request handler.

    A filter is either a coroutine function or an asynchronous generator.

    A coroutine function filter can inspect and modify a request, and:
    • return no value to indicate that the filter passed; processing continues down the chain
    • return a response; the request is not processed by any subsequent filters or handler

    An asynchronous generator filter can inspect and modify a request, and:
    • yield no value; processing continues down the chain
    • yield a response; the request is not processed by any subsequent filters or handler

    The response from handler and downstream filters is sent to the asynchronous generator
    filter and becomes the result of the yield expression. The filter can then inspect and
    modify the response, and:
    • yield no value; the existing response is passed back up the chain
    • yield a new response; this response is passed back up the chain to the caller
    """

    def __init__(self, *, filters: MutableSequence[Callable] | None = None, handler: Callable):
        """Initialize a filter chain."""
        self.filters = filters  # concrete and mutable
        self.handler = handler

    async def __call__(self, request: Request) -> Response:
        """Handle a request."""
        unwind = []
        response = None
        exception = None
        for filter in (f(request) for f in self.filters):
            if inspect.isasyncgen(filter):
                try:
                    response = await filter.__anext__()
                    if not response:
                        unwind.append(filter)
                except StopAsyncIteration:
                    pass
                except Exception as e:
                    exception = e
            elif asyncio.iscoroutine(filter):
                try:
                    response = await filter
                except Exception as e:
                    exception = e
            if exception or response:
                break
        if not exception and not response:
            try:
                response = await self.handler(request)
            except Exception as e:
                exception = e
        for filter in reversed(unwind):
            try:
                _response = (
                    await filter.asend(response)
                    if not exception
                    else await filter.athrow(
                        type(exception), exception, exception.__traceback__
                    )
                )
                if _response:  # new response overrides previous response or exception
                    response = _response
                    exception = None
            except StopAsyncIteration:
                pass
            except Exception as e:  # new exception overrides previous response or exception
                exception = e
        if exception:
            raise exception
        return response


BasicCredentials = namedtuple("BasicCredentials", ("user_id", "password"))


class BasicScheme(Scheme):
    """
    HTTP basic authentication scheme.

    Parameters:
    • name: name of authentication scheme
    • description: a short description of authentication scheme
    """

    def extract(self, request: Request) -> BasicCredentials | None:
        """Return basic credentials from request if provided, otherwise None."""
        try:
            scheme, credentials = request.headers["Authorization"].split(" ", 1)
            if scheme.lower() == "basic":
                return BasicCredentials(*base64.b64decode(credentials).decode().split(":", 1))
        except:
            return None


class BearerScheme(Scheme):
    """
    Bearer token authentication scheme.

    Parameters:
    • name: name of authentication scheme
    • description: a short description of authentication scheme
    • format: identifies how bearer token is formatted
    """

    def __init__(self, *, format: str | None = None, **kwargs):
        super().__init__(**kwargs)
        self.format = format

    def extract(self, request: Request) -> str | None:
        """Return bearer token value from request if provided, otherwise None."""
        try:
            name, token = request.headers["Authorization"].split(" ", 1)
            if name.lower() == "bearer":
                return token
        except:
            pass
        return None


class CookieScheme(Scheme):
    """
    Cookie authentication scheme.

    Parameters:
    • name: name of authentication scheme
    • description: a short description of authentication scheme
    • cookie: name of cookie to be used
    """

    def __init__(self, *, cookie: str, **kwargs):
        super().__init__(**kwargs)
        self.cookie = cookie

    def extract(self, request: Request) -> str | None:
        """Return cookie value from request if provided, otherwise None."""
        try:
            return request.cookies[self.cookie].value
        except:
            return None


class HeaderScheme(Scheme):
    """
    Header authentication scheme.

    Parameters:
    • name: name of authentication scheme
    • description: a short description of authentication scheme
    • header: name of the header to be used
    """

    def __init__(self, *, header: str, **kwargs):
        super().__init__(**kwargs)
        self.header = header

    def extract(self, request: Request) -> str | None:
        """Return header value from request if provided, otherwise None."""
        try:
            return request.headers[self.header]
        except:
            return None


async def _subordinate(resource, segment: str):

    # resource.attr
    if (attr := getattr(resource, segment, None)) is not None:
        if fondat.resource.is_resource(attr):
            return attr
        if not callable(attr):
            raise NotFoundError
        hints = typing.get_type_hints(attr)
        if not fondat.resource.is_resource(hints.get("return")):
            raise NotFoundError
        if asyncio.iscoroutinefunction(attr):
            return await attr()
        else:
            return attr()

    # resource[item]
    try:
        hints = typing.get_type_hints(resource.__getitem__)
        name = next(iter(hints))
        if name == "return":
            raise NotFoundError
        item = resource[StringCodec.get(hints[name]).decode(segment)]
        if not fondat.resource.is_resource(item):
            raise NotFoundError
        return item
    except:
        raise NotFoundError


class InQuery:
    """
    Annotation to indicate an operation parameter is expected in a request query string
    parameter. This is the default annotation for query operation parameters.

    If the InQuery class is used as the annotation instead of an InQuery(name=...) instance,
    then the name of the query string parameter will be the name of the operation parameter.

    Parameters:
    • name: name of the query string parameter
    """

    __slots__ = {"name"}

    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return f"request query string parameter: {self.name}"

    def __repr__(self):
        return f"InQuery({self.name})"


class InBody:
    """
    Annotation to indicate an operation parameter is expected in a body parameter. This is the
    default annotation for mutation operation parameters.

    If the InBody class is used as the annotation instead of an InBody(name=...) instance,
    then the name of the body parameter will be the name of the operation parameter.

    Parameters:
    • name: name of the body parameter
    """

    __slots__ = {"name"}

    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return f"request body parameter: {self.name}"

    def __repr__(self):
        return f"InBody({self.name})"


class AsBody:
    """Annotation to indicate an operation parameter is expected to be the request body."""

    def __str__(self):
        return f"request body"

    def __repr__(self):
        return "AsBody"


def get_param_in(method, param_name, type_hint):
    """
    Return an annotation expressing where a parameter is to be provided.

    If an annotations is a class, then it is instantated. If no annotation exists, then an
    appropriate InQuery annotation is provided.
    """
    stripped = fondat.types.strip_optional(type_hint)
    if typing.get_origin(stripped) is Annotated:
        args = typing.get_args(stripped)
        for annotation in args[1:]:
            if is_subclass(annotation, (InBody, InQuery)):
                return annotation(param_name)
            elif is_subclass(annotation, AsBody):
                return annotation()
            elif isinstance(annotation, (AsBody, InBody, InQuery)):
                return annotation
    if method._fondat_operation.type == "mutation":
        return InBody(param_name)
    return InQuery(param_name)


@functools.cache
def get_body_type(operation):
    """Return the type of the request body for the specified operation."""
    signature = inspect.signature(operation)
    type_hints = typing.get_type_hints(operation, include_extras=True)
    as_body_param = None
    in_body_params = {}
    required_keys = set()
    for name, hint in type_hints.items():
        if name == "return":
            continue
        param_in = get_param_in(operation, name, hint)
        if not isinstance(param_in, (AsBody, InBody)):
            continue
        is_required = signature.parameters[name].default is inspect.Parameter.empty
        match param_in:
            case InBody():
                in_body_params[param_in.name] = hint
                if is_required:
                    required_keys.add(param_in.name)
            case AsBody():
                if as_body_param:
                    raise TypeError("multiple AsBody annotated parameters")
                as_body_param = name
    if as_body_param and in_body_params:
        raise TypeError("mixed AsBody and InBody annotated parameters")
    if as_body_param:
        return type_hints[as_body_param]
    if not in_body_params:
        return None
    rb = TypedDict("RequestBody", in_body_params, total=False)
    rb.__required_keys__ = frozenset(required_keys)
    rb.__optional_keys__ = frozenset(k for k in in_body_params if k not in required_keys)
    return rb


async def simple_error_filter(request: Request):
    """Generates a simple JSON error response if an exception is raised."""
    try:
        try:
            yield
        except fondat.error.Error:
            raise
        except Exception as e:
            _logger.exception("unhandled exception")
            raise InternalServerError from e
    except fondat.error.Error as err:
        body = str(err)
        response = Response()
        response.status = err.status
        response.headers["content-type"] = "text/plain; charset=UTF-8"
        response.headers["content-length"] = str(len(body))
        response.body = BytesStream(body.encode())
        yield response


async def _decode_body(operation, request):
    body_type = get_body_type(operation)
    if not body_type:
        return None
    python_type = fondat.types.strip_annotations(body_type)
    if is_subclass(python_type, Stream):
        return request.body
    async with request.body as stream:  # close stream after reading
        content = await read_stream(stream)
    if len(content) == 0:
        return None  # empty body is no body
    try:
        with DecodeError.path_on_error("«body»"):
            result = BinaryCodec.get(body_type).decode(content)
    except DecodeError as de:
        raise BadRequestError from de
    except Exception as e:
        raise InternalServerError from e
    return result


class Application:
    """
    An HTTP application, which handles ncoming HTTP requests by:
    • passing request through zero or more HTTP filters, and
    • dispatching it to a resource.

    Parameters and attributes:
    • root: resource to dispatch requests to
    • filters: filters to apply during HTTP request processing
    • path: URI path to root resource

    An HTTP application is a request handler; it's a coroutine callable that handles an HTTP
    request and returns an HTTP response. For a description of filters, see: Chain.
    """

    def __init__(
        self,
        root: Any,
        *,
        filters: Iterable[Any] = (simple_error_filter,),
        path: str = "/",
    ):
        if not fondat.resource.is_resource(root):
            raise TypeError("root is not a resource")
        self.root = root
        self.path = path.rstrip("/") + "/"
        self.filters = list(filters or [])

    async def __call__(self, request: Request) -> Response:
        return await Chain(filters=self.filters, handler=self._handle)(request)

    async def _handle(self, request: Request) -> Response:
        if not request.path.startswith(self.path):
            raise NotFoundError
        path = request.path[len(self.path) :]
        response = Response()
        method = request.method.lower()
        segments = path.split("/") if path else ()
        resource = self.root
        operation = None
        for segment in segments:
            if operation:  # cannot have segments after operation name
                raise NotFoundError
            try:
                resource = await _subordinate(resource, segment)
            except NotFoundError:
                try:
                    operation = getattr(resource, segment)
                    if not fondat.resource.is_operation(operation):
                        raise NotFoundError
                except AttributeError:
                    raise NotFoundError
        if operation:  # operation name as segment (@query or @mutation)
            fondat_op = getattr(operation, "_fondat_operation", None)
            if not fondat_op or not fondat_op.method == method:
                raise MethodNotAllowedError
        else:  # no remaining segments; operation name as HTTP method
            operation = getattr(resource, method, None)
            if not fondat.resource.is_operation(operation):
                raise MethodNotAllowedError
        body = await _decode_body(operation, request)
        params = {}
        signature = inspect.signature(operation)
        hints = typing.get_type_hints(operation, include_extras=True)
        return_hint = hints.get("return", type(None))
        for name, hint in hints.items():
            if name == "return":
                continue
            required = signature.parameters[name].default is inspect.Parameter.empty
            param_in = get_param_in(operation, name, hint)
            match param_in:
                case AsBody() if body is not None:
                    params[name] = body
                case InBody() if body is not None:
                    if param_in.name in body:
                        params[name] = body[param_in.name]
                case InQuery() if param_in.name in request.query:
                    codec = StringCodec.get(hint)
                    try:
                        with DecodeError.path_on_error(param_in.name):
                            params[name] = codec.decode(request.query[param_in.name])
                    except DecodeError as de:
                        raise BadRequestError from de
            if name not in params and required:
                if not is_optional(hint):
                    raise BadRequestError from DecodeError(
                        "required parameter", ["«params»", name]
                    )
                params[name] = None
        result = await operation(**params)
        if not is_subclass(return_hint, Stream):
            return_codec = BinaryCodec.get(return_hint)
            try:
                result = BytesStream(return_codec.encode(result), return_codec.content_type)
            except Exception as e:
                raise InternalServerError from e
        response.body = result
        response.headers["Content-Type"] = response.body.content_type
        if response.body.content_length is not None:
            if response.body.content_length == 0:
                response.status = http.HTTPStatus.NO_CONTENT.value
            else:
                response.headers["Content-Length"] = str(response.body.content_length)
        return response



"""Module for authentication and authorization of resource operations."""

from collections.abc import Callable, Coroutine, Iterable
from typing import Any


class Scheme:
    """Base class for authentication scheme."""

    def __init__(self, *, name: str, description: str | None = None):
        self.name = name
        self.description = description


class Policy:
    """
    A combination of authentication schemes and authorization rules. For a security policy to
    allow an operation to be performed, all authentication schemes must be satisfied and all
    authorization rules must pass.

    Parameters and attributes:
    • schemes: authentication schemes that must be satisfied
    • rules: authorization rules that must pass

    If schemes is None, then authenticaton is not applicable. If schemes is empty, then the
    policy allows access without authentication.

    An authorization rule is a coroutine function that raises a security exception if
    authorization to an operation should not granted:

    • UnauthorizedError: user could not be authenticated (misnomer)
    • ForbiddenError: user is authenticated and is denied access
    """

    __slots__ = {"schemes", "rules"}

    def __init__(
        self,
        schemes: Iterable[Scheme] | None = None,
        rules: Iterable[Callable[[], Coroutine[Any, Any, Any]]] | None = None,
    ):
        self.schemes = schemes
        self.rules = rules or ()

    async def apply(self):
        """
        Apply the security policy by evaluating the authorization rules.

        When a security policy is applied, authorization rules are evaluated in the order
        specified. The first exception encountered is raised immediately, ceasing further
        evaluation.
        """
        for rule in self.rules:
            await rule()



"""Module to expose resources through ASGI."""

import fondat.http
import logging
import urllib.parse

from collections.abc import Awaitable, Callable, Mapping
from fondat.error import InternalServerError
from fondat.stream import Stream


_logger = logging.getLogger(__name__)


def _int(s: str) -> int:
    result = None
    if s is not None:
        try:
            return int(s)
        except:
            pass
    return result


class ReceiveStream(Stream):
    """Stream that encapsulates the ASGI receive interface."""

    def __init__(self, scope: Mapping, receive: Awaitable):
        for key, value in scope.get("headers"):
            if key == "content-type":
                self.content_type = value
            elif key == "content-length":
                self.content_length = _int(value)
        self._receive = receive
        self._more = True

    async def __anext__(self) -> bytes:
        if not self._more:
            raise StopAsyncIteration
        event = await self._receive()
        event_type = event["type"]
        if event_type == "http.disconnect":
            raise StopAsyncIteration
        if event_type != "http.request":
            raise InternalServerError from ValueError(
                "expecting http.request event type; received {event_type}"
            )
        self._more = event.get("more_body", False)
        return event.get("body", b"")

    async def close(self):
        self._more = False


def asgi_app(
    handler: Callable, startup: Callable | None = None, shutdown: Callable | None = None
) -> Callable:
    """
    Expose a Fondat HTTP request handler as an ASGI application.

    Parameters:
    • handler: HTTP handler coroutine function
    • startup: lifespan startup coroutine function
    • shutdown: lifespan shutdown coroutine function

    The HTTP handler coroutine function is called in response to ASGI HTTP protocol events.

    The startup and shutdown coroutine functions are called in response to ASGI lifespan
    protocol events. This allows the application to initialize and shutdown in the context of
    a running event loop.
    """

    async def lifespan(scope, receive, send):
        message = await receive()
        lifespan_type = message["type"]
        try:
            match lifespan_type:
                case "lifespan.startup":
                    if startup is not None:
                        await startup()
                case "lifespan.shutdown":
                    if shutdown is not None:
                        await shutdown()
                case _:
                    _logger.warn("ignoring ASGI lifespan type: {lifespan_type}")
        except Exception as e:
            _logger.exception(f"ASGI {lifespan_type} failed")
            await send({"type": f"{lifespan_type}.failed", "message": str(e)})
            raise
        await send({"type": f"{lifespan_type}.complete"})

    async def http(scope, receive, send):
        request = fondat.http.Request()
        request.method = scope["method"]
        request.path = scope["path"]
        request.version = scope["http_version"]
        for key, value in scope["headers"]:
            request.headers.add(key.decode(), value.decode())
        for header in request.headers.popall("cookie", ()):
            request.cookies.load(header)
        request.query = fondat.http.Query(
            urllib.parse.parse_qsl((scope.get("query_string") or b"").decode())
        )
        request.body = ReceiveStream(scope, receive)
        response = await handler(request)
        headers = ((k.lower().encode(), v.encode()) for k, v in response.headers.items())
        headers = [*headers]
        cookies = (
            (
                (k.lower().encode(), v.encode())
                for k, v in (
                    header.split(": ")
                    for header in response.cookies.output(sep="\n").split("\n")
                )
            )
            if response.cookies
            else ()
        )
        cookies = [*cookies]
        await send(
            {
                "type": "http.response.start",
                "status": response.status,
                "headers": [*headers, *cookies],
            }
        )
        if response.body is not None:
            async with response.body:  # close stream after reading
                async for chunk in response.body:
                    await send(
                        {
                            "type": "http.response.body",
                            "body": chunk,
                            "more_body": True,
                        }
                    )
        await send(
            {
                "type": "http.response.body",
                "more_body": False,
            }
        )

    async def app(scope, receive, send):
        """Coroutine that implements ASGI interface."""
        scope_type = scope["type"]
        match scope_type:
            case "http":
                return await http(scope, receive, send)
            case "lifespan":
                return await lifespan(scope, receive, send)
            case _:
                _logger.warn(f"unrecognized ASGI scope type: {scope_type}")

    return app



"""Module to support encoding and decoding of values."""

import base64
import csv
import dataclasses
import fondat.types
import io
import iso8601
import json
import keyword
import logging
import typing

from collections import namedtuple
from collections.abc import Iterable, Mapping, Set
from contextlib import contextmanager, suppress
from datetime import date, datetime, timezone
from decimal import Decimal
from fondat.types import is_optional, is_subclass, strip_annotations
from types import NoneType, UnionType
from typing import Any, Generic, Literal, TypeVar, Union, get_args, get_origin
from uuid import UUID


_logger = logging.getLogger(__name__)


# ----- content types -----


APPLICATION_JSON = "application/json"
APPLICATION_OCTET_STREAM = "application/octet-stream"
TEXT_PLAIN = "text/plain; charset=UTF-8"


# ----- type aliases -----


BinaryType = bytes | bytearray
JSONType = Any
StringType = str


# ----- utilities -----


@contextmanager
def _wrap(exception):
    try:
        yield
    except Exception as e:
        if isinstance(e, exception):
            raise
        raise exception from e


def _csv_encode(value):
    sio = io.StringIO()
    csv.writer(sio).writerow(value)
    return sio.getvalue().rstrip("\r\n")


def _csv_decode(value):
    if not isinstance(value, str):
        raise DecodeError
    return csv.reader([value]).__next__()


def _b2s(b):
    if not isinstance(b, bytes | bytearray):
        raise DecodeError
    with _wrap(DecodeError):
        return b.decode()


def _s2j(s):
    if not isinstance(s, str):
        raise DecodeError
    with _wrap(DecodeError):
        return json.loads(s)


# ----- errors -----


class CodecError(ValueError):
    """
    Error raised in the event that a value cannot be decoded.
    """

    __slots__ = {"message", "path"}

    def __init__(self, message: str | None = None, path: list[str | int] | None = None):
        self.message = message
        self.path = path

    def __repr__(self):
        return f"{self.__class__.__name__}({self.message!r}, {self.path!r})"

    def __str__(self):
        return " ".join(str(s) for s in (self.message, self.path) if s is not None)

    @staticmethod
    @contextmanager
    def path_on_error(path: list[str | int] | str | int) -> None:
        """Context manager to add to error path in the event that a DecodeError is raised."""
        try:
            yield
        except CodecError as ce:
            if ce.path is None:
                ce.path = []
            match path:
                case str() | int():
                    ce.path.insert(0, path)
                case list():
                    ce.path = path + ce.path
            raise


class EncodeError(CodecError):
    """..."""


class DecodeError(CodecError):
    """..."""


# ----- base -----


PT = TypeVar("PT")  # Python type hint
TT = TypeVar("TT")  # target type hint


class Codec(Generic[PT, TT]):
    """
    Base class for all things encode and decode.
    """

    def __init__(self, python_type: Any):
        self.python_type = python_type

    @staticmethod
    def handles(python_type: Any) -> bool:
        """Return True if the codec handles the specified Python type."""
        raise NotImplementedError

    @classmethod
    def get(cls, python_type: Any) -> "Codec[PT, TT]":
        """
        Return a codec that handles the specified Python type.

        If the subclass contains a `_cache` attribute, and does not

        """
        if cls is Codec:
            raise NotImplementedError
        with suppress(AttributeError, KeyError):
            return cls._cache[python_type]
        for codec_class in cls.__subclasses__():
            if codec_class.handles(python_type):
                codec = codec_class(python_type)
                with suppress(AttributeError):
                    cache = getattr(codec, "_cache", False)
                    if isinstance(cache, Mapping):
                        cache[python_type] = codec
                return codec
        raise TypeError(f"no codec for {python_type}")

    def encode(self, value: PT) -> TT:
        """Encode value from F type to T type."""
        raise NotImplementedError

    def decode(self, value: TT) -> PT:
        """Decode value from T type to F type."""
        raise NotImplementedError


class StringCodec(Codec[PT, StringType]):
    """Encodes Python types to/from Unicode string representations."""

    _cache = {}

    def encode(self, value: PT) -> StringType:
        """Encode value from Python type to string type."""
        raise NotImplementedError

    def decode(self, value: StringType) -> PT:
        """Decode value from string type to Python type."""
        raise NotImplementedError


class BinaryCodec(Codec[PT, BinaryType]):
    """Encodes Python types to/from binary representations.

    Attribute:
    • content_type: string containing the media type of the binary representation
    """

    content_type = APPLICATION_OCTET_STREAM

    _cache = {}

    def encode(self, value: PT) -> BinaryType:
        """Encode value from Python type to binary type."""
        raise NotImplementedError

    def decode(self, value: BinaryType) -> PT:
        """Decode value from binary type to Python type."""
        raise NotImplementedError


class JSONCodec(Codec[PT, JSONType]):
    """Encodes Python types to/from the JSON representations."""

    _cache = {}

    def encode(self, value: PT) -> JSONType:
        """Encode value from Python type to binary type."""
        raise NotImplementedError

    def decode(self, value: JSONType) -> PT:
        """Decode value from binary type to Python type."""
        raise NotImplementedError


# ----- str -----


class StrBinaryCodec(BinaryCodec[str]):
    """Bytes codec for Unicode character strings."""

    content_type = TEXT_PLAIN

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, str)

    def encode(self, value: str) -> BinaryType:
        if not isinstance(value, str):
            raise EncodeError
        with _wrap(EncodeError):
            return value.encode()

    def decode(self, value: BinaryType) -> str:
        return _b2s(value)


class StrStringCodec(StringCodec[str]):
    """String codec for Unicode character strings."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, str)

    def encode(self, value: str) -> StringType:
        if not isinstance(value, str):
            raise EncodeError
        return value

    def decode(self, value: StringType) -> str:
        if not isinstance(value, str):
            raise DecodeError
        return value


class StrJSONCodec(JSONCodec[str]):
    """JSON codec for Unicode character strings."""

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = StrStringCodec(python_type)

    @staticmethod
    def handles(python_type: Any) -> bool:
        return StrStringCodec.handles(python_type)

    def encode(self, value: str) -> JSONType:
        return self.codec.encode(value)

    def decode(self, value: JSONType) -> str:
        if not isinstance(value, str):
            raise DecodeError
        return self.codec.decode(value)


# ----- bytes/bytearray -----


class BytesBinaryCodec(BinaryCodec[bytes | bytearray]):
    """Binary codec for byte sequences."""

    content_type = APPLICATION_OCTET_STREAM

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, bytes | bytearray)

    def encode(self, value: bytes | bytearray) -> BinaryType:
        if not isinstance(value, bytes | bytearray):
            raise EncodeError
        return value

    def decode(self, value: BinaryType) -> bytes | bytearray:
        if not isinstance(value, BinaryType):
            raise DecodeError
        return value


class BytesStringCodec(StringCodec[bytes | bytearray]):
    """
    String codec for byte sequences. A byte sequence is represented in string values as a
    base64-encoded string. Example: "SGVsbG8gRm9uZGF0".
    """

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, BinaryType)

    def encode(self, value: bytes | bytearray) -> StringType:
        with _wrap(EncodeError):
            return base64.b64encode(value).decode()

    def decode(self, value: StringType) -> bytes | bytearray:
        with _wrap(DecodeError):
            return base64.b64decode(value)


class BytesJSONCodec(JSONCodec[bytes | bytearray]):
    """
    JSON codec for byte sequences. A byte sequence is represented in JSON values as a
    base64-encoded string. Example: "SGVsbG8gRm9uZGF0".
    """

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = BytesStringCodec(python_type)

    @staticmethod
    def handles(python_type: Any) -> bool:
        return BytesStringCodec.handles(python_type)

    def encode(self, value: JSONType) -> bytes | bytearray:
        return self.codec.encode(value)

    def decode(self, value: bytes | bytearray) -> JSONType:
        return self.codec.decode(value)


# ----- int -----


class IntStringCodec(StringCodec[int]):
    """String codec for integers."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, int) and not is_subclass(python_type, bool)

    def encode(self, value: int) -> StringType:
        if not isinstance(value, int) or isinstance(value, bool):
            raise EncodeError
        return str(value)

    def decode(self, value: StringType) -> int:
        if not isinstance(value, str):
            raise DecodeError
        with _wrap(DecodeError):
            return int(value)


class IntBinaryCodec(BinaryCodec[int]):
    """Binary codec for integers."""

    content_type = TEXT_PLAIN

    @staticmethod
    def handles(python_type: Any) -> bool:
        return IntStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = IntStringCodec(python_type)

    def encode(self, value: int) -> BinaryType:
        return self.codec.encode(value).encode()

    def decode(self, value: BinaryType) -> int:
        return self.codec.decode(_b2s(value))


class IntJSONCodec(JSONCodec[int]):
    """JSON codec for integers."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, int) and not is_subclass(python_type, bool)

    def encode(self, value: int) -> JSONType:
        if not isinstance(value, int) or isinstance(value, bool):
            raise EncodeError
        return value

    def decode(self, value: JSONType) -> int:
        if not isinstance(value, int | float) or isinstance(value, bool):
            raise DecodeError
        result = value
        if isinstance(result, float):
            result = int(result)
            if result != value:  # 1.0 == 1
                raise DecodeError
        return result


# ----- float -----


class FloatStringCodec(StringCodec[float]):
    """String codec for floating point numbers."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, float)

    def encode(self, value: float) -> StringType:
        if not isinstance(value, float):
            raise EncodeError
        return str(value)

    def decode(self, value: StringType) -> float:
        if not isinstance(value, str):
            raise DecodeError
        with _wrap(DecodeError):
            return float(value)


class FloatBinaryCodec(BinaryCodec[float]):
    """Binary codec for floating point numbers."""

    content_type = TEXT_PLAIN

    @staticmethod
    def handles(python_type: Any) -> bool:
        return FloatStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = FloatStringCodec(python_type)

    def encode(self, value: float) -> BinaryType:
        with _wrap(EncodeError):
            return self.codec.encode(value).encode()

    def decode(self, value: BinaryType) -> float:
        return self.codec.decode(_b2s(value))


class FloatJSONCodec(JSONCodec[float]):
    """JSON codec for floating point numbers."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, float)

    def encode(self, value: float) -> JSONType:
        if not isinstance(value, float):
            raise EncodeError
        return value

    def decode(self, value: JSONType) -> float:
        if not isinstance(value, int | float) or isinstance(value, bool):
            raise DecodeError
        return float(value)


# ----- bool -----


class BoolStringCodec(StringCodec[bool]):
    """String codec for boolean values."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, bool)

    def encode(self, value: bool) -> StringType:
        if not isinstance(value, bool):
            raise EncodeError
        return "true" if value else "false"

    def decode(self, value: StringType) -> bool:
        if not isinstance(value, str):
            raise DecodeError
        try:
            return {"true": True, "false": False}[value]
        except KeyError:
            raise DecodeError


class BoolBinaryCodec(BinaryCodec[bool]):
    """Binary codec for boolean values."""

    content_type = TEXT_PLAIN

    @staticmethod
    def handles(python_type: Any) -> bool:
        return BoolStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = BoolStringCodec(python_type)

    def encode(self, value: bool) -> BinaryType:
        with _wrap(EncodeError):
            return self.codec.encode(value).encode()

    def decode(self, value: BinaryType) -> bool:
        return self.codec.decode(_b2s(value))


class BoolJSONCodec(JSONCodec[bool]):
    """JSON codec for boolean values."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, bool)

    def encode(self, value: bool) -> JSONType:
        if not isinstance(value, bool):
            raise EncodeError
        return value

    def decode(self, value: JSONType) -> bool:
        if not isinstance(value, bool):
            raise DecodeError
        return value


# ----- NoneType -----


class NoneTypeStringCodec(StringCodec[NoneType]):
    """String codec for None value."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return python_type is NoneType

    def encode(self, value: NoneType) -> StringType:
        if value is not None:
            raise EncodeError
        return ""

    def decode(self, value: StringType) -> NoneType:
        if value != "":
            raise DecodeError
        return None


class NoneTypeBinaryCodec(BinaryCodec[NoneType]):
    """Binary codec for None value."""

    content_type = TEXT_PLAIN

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return python_type is NoneType

    def encode(self, value: NoneType) -> BinaryType:
        if value is not None:
            raise EncodeError
        return b""

    def decode(self, value: BinaryType) -> NoneType:
        if value != b"":
            raise DecodeError
        return None


class NoneTypeJSONCodec(JSONCodec[NoneType]):
    """JSON codec for None value."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return python_type is NoneType

    def encode(self, value: NoneType) -> JSONType:
        if value is not None:
            raise EncodeError
        return None

    def decode(self, value: JSONType) -> NoneType:
        if value is not None:
            raise DecodeError
        return None


# ----- Decimal -----


class DecimalStringCodec(StringCodec[Decimal]):
    """String codec for Decimal numbers."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, Decimal)

    def encode(self, value: Decimal) -> StringType:
        if not isinstance(value, Decimal):
            raise EncodeError
        return str(value)

    def decode(self, value: StringType) -> Decimal:
        if not isinstance(value, str):
            raise DecodeError
        with _wrap(DecodeError):
            return Decimal(value)


class DecimalBinaryCodec(BinaryCodec[Decimal]):
    """Binary codec for Decimal numbers."""

    content_type = TEXT_PLAIN

    @staticmethod
    def handles(python_type: Any) -> bool:
        return DecimalStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = DecimalStringCodec(python_type)

    def encode(self, value: Decimal) -> BinaryType:
        return self.codec.encode(value).encode()

    def decode(self, value: BinaryType) -> Decimal:
        return self.codec.decode(_b2s(value))


class DecimalJSONCodec(JSONCodec[Decimal]):
    """
    JSON codec for Decimal numbers. Decimal numbers are represented in JSON as strings, due to
    the imprecision of floating point numbers.
    """

    @staticmethod
    def handles(python_type: Any) -> bool:
        return DecimalStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = DecimalStringCodec(python_type)

    def encode(self, value: Decimal) -> JSONType:
        return self.codec.encode(value)

    def decode(self, value: JSONType) -> Decimal:
        return self.codec.decode(value)


# ----- date -----


class DateStringCodec(StringCodec[date]):
    """
    String codec for dates. A date is represented in a string in RFC 3339 format.
    Example: "2018-06-16".
    """

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, date) and not is_subclass(python_type, datetime)

    def encode(self, value: date) -> StringType:
        if not isinstance(value, date):
            raise EncodeError
        return value.isoformat()

    def decode(self, value: StringType) -> date:
        if not isinstance(value, str):
            raise DecodeError
        with _wrap(DecodeError):
            return date.fromisoformat(value)


class DateBinaryCodec(BinaryCodec[date]):
    """
    Binary codec for dates. A date is represented in a binary format as an RFC 3339 UTF-8 or
    ASCII encoded string. Example: "2018-06-16".
    """

    content_type = TEXT_PLAIN

    @staticmethod
    def handles(python_type: Any) -> bool:
        return DateStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = DateStringCodec(python_type)

    def encode(self, value: date) -> BinaryType:
        return self.codec.encode(value).encode()

    def decode(self, value: BinaryType) -> date:
        return self.codec.decode(_b2s(value))


class DateJSONCodec(JSONCodec[date]):
    """
    JSON codec for dates. A date is represented in JSON as an RFC 3339 formatted string.
    Example: "2018-06-16".
    """

    @staticmethod
    def handles(python_type: Any) -> bool:
        return DateStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = DateStringCodec(python_type)

    def encode(self, value: date) -> JSONType:
        return self.codec.encode(value)

    def decode(self, value: JSONType) -> date:
        return self.codec.decode(value)


# ----- datetime -----


def _to_utc(value):
    if value.tzinfo is None:  # naive value interpreted as UTC
        value = value.replace(tzinfo=timezone.utc)
    return value.astimezone(timezone.utc)


class DatetimeStringCodec(StringCodec[datetime]):
    """
    String codec for datetime.

    It will decode a datetime represented in an ISO 8601 formatted string. It will encode a
    datetime to an RFC 3339 (subset of ISO 8601) formatted string.

    Datetimes always encode and decode to UTC timezone offset.

    Example: "2020-04-07T12:34:56.789012Z".
    """

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, datetime)

    def encode(self, value: datetime) -> StringType:
        if not isinstance(value, datetime):
            raise EncodeError
        result = _to_utc(value).isoformat()
        if result.endswith("+00:00"):
            result = result[0:-6]
        if "+" not in result and not result.endswith("Z"):
            result = f"{result}Z"
        return result

    def decode(self, value: StringType) -> datetime:
        if not isinstance(value, str):
            raise DecodeError
        with _wrap(DecodeError):
            return _to_utc(iso8601.parse_date(value))


class DatetimeBinaryCodec(BinaryCodec[datetime]):
    """
    Binary codec for datetime.

    It will decode a datetime represented in an ISO 8601 formatted UTF-8 or ASCII encoded byte
    string. It will encode a datetime to an RFC 3339 (subset of ISO 8601) formatted UTF-8
    encoded byte string.

    Datetimes always encode and decode to UTC timezone offset.

    Example: b"2020-04-07T12:34:56.789012Z".
    """

    content_type = TEXT_PLAIN

    @staticmethod
    def handles(python_type: Any) -> bool:
        return DatetimeStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = DatetimeStringCodec(python_type)

    def encode(self, value: datetime) -> BinaryType:
        return self.codec.encode(value).encode()

    def decode(self, value: BinaryType) -> datetime:
        return self.codec.decode(_b2s(value))


class DatetimeJSONCodec(JSONCodec[datetime]):
    """
    String codec for datetime.

    It will decode a datetime represented in an ISO 8601 formatted string. It will encode a
    datetime to an RFC 3339 (subset of ISO 8601) formatted string.

    Datetimes always encode and decode to UTC timezone offset.

    Example: "2020-04-07T12:34:56.789012Z".
    """

    @staticmethod
    def handles(python_type: Any) -> bool:
        return DatetimeStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = DatetimeStringCodec(python_type)

    def encode(self, value: datetime) -> JSONType:
        return self.codec.encode(value)

    def decode(self, value: JSONType) -> datetime:
        return self.codec.decode(value)


# ----- UUID -----


class UUIDStringCodec(StringCodec[UUID]):
    """String codec for UUID."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, UUID)

    def encode(self, value: UUID) -> StringType:
        if not isinstance(value, UUID):
            raise EncodeError
        return str(value)

    def decode(self, value: StringType) -> UUID:
        if not isinstance(value, str):
            raise DecodeError
        with _wrap(DecodeError):
            return UUID(value)


class UUIDBinaryCodec(BinaryCodec[UUID]):
    """Binary codec for UUID."""

    content_type = TEXT_PLAIN

    @staticmethod
    def handles(python_type: Any) -> bool:
        return UUIDStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = UUIDStringCodec(python_type)

    def encode(self, value: UUID) -> BinaryType:
        return self.codec.encode(value).encode()

    def decode(self, value: BinaryType) -> UUID:
        return self.codec.decode(_b2s(value))


class UUIDJSONCodec(JSONCodec[UUID]):
    """JSON codec for UUID."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        return UUIDStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = UUIDStringCodec(python_type)

    def encode(self, value: UUID) -> JSONType:
        return self.codec.encode(value)

    def decode(self, value: JSONType) -> UUID:
        return self.codec.decode(value)


# ----- TypedDict -----


class TypedDictJSONCodec(JSONCodec[PT]):
    """..."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return typing.is_typeddict(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        python_type = strip_annotations(python_type)
        self.hints = typing.get_type_hints(python_type, include_extras=True)
        if {type(k) for k in self.hints.keys()} != {str}:
            raise TypeError("codec only supports TypedDict with str keys")

    def _process(self, value: dict[str, Any], method) -> dict[str, Any]:
        result = {}
        for key in self.hints:
            codec = JSONCodec.get(self.hints[key])
            with suppress(KeyError):
                with CodecError.path_on_error(key):
                    result[key] = getattr(codec, method)(value[key])
        return result

    def encode(self, value: PT) -> JSONType:
        if not isinstance(value, dict):
            raise EncodeError
        return self._process(value, "encode")

    def decode(self, value: JSONType) -> PT:
        return self._process(value, "decode")


class TypedDictStringCodec(StringCodec[PT]):
    """..."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        return TypedDictJSONCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = TypedDictJSONCodec(python_type)

    def encode(self, value: PT) -> StringType:
        if not isinstance(value, dict):
            raise EncodeError
        return json.dumps(self.codec.encode(value))

    def decode(self, value: StringType) -> PT:
        return self.codec.decode(_s2j(value))


class TypedDictBinaryCodec(BinaryCodec[PT]):
    """..."""

    content_type = APPLICATION_JSON

    @staticmethod
    def handles(python_type: Any) -> bool:
        return TypedDictStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = TypedDictStringCodec(python_type)

    def encode(self, value: PT) -> BinaryType:
        return self.codec.encode(value).encode()

    def decode(self, value: BinaryType) -> PT:
        return self.codec.decode(_b2s(value))


# ----- tuple -----


class _TupleCodec(Codec[PT, TT]):
    @classmethod
    def handles(cls, python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return is_subclass(python_type, tuple) or is_subclass(get_origin(python_type), tuple)

    def __init__(self, python_type: Any, base_codec_type: type[Codec[PT, TT]]):
        python_type = strip_annotations(python_type)
        args = get_args(python_type) or (Any, ...)
        if len(args) != 2 and Ellipsis in args or args[0] is Ellipsis:
            raise TypeError(f"unexpected ellipsis in tuple[{', '.join(args)}]")
        self.varg = args[0] if len(args) == 2 and args[1] is Ellipsis else None
        self.args = () if self.varg else args
        self.codecs = [base_codec_type.get(arg) for arg in self.args]
        self.vcodec = base_codec_type.get(self.varg) if self.varg else None

    def _encode(self, value: PT) -> list[Any]:
        if not isinstance(value, tuple) or (self.args and len(value) != len(self.args)):
            raise EncodeError
        # TODO: path
        return list(
            (self.vcodec.encode(item) for item in value)
            if self.vcodec
            else (self.codecs[n].encode(value[n]) for n in range(len(self.codecs)))
        )

    def _decode(self, value: list[Any]) -> PT:
        if not isinstance(value, list) or (self.args and len(value) != len(self.args)):
            raise DecodeError
        # TODO: path
        return tuple(
            (self.vcodec.decode(item) for item in value)
            if self.vcodec
            else (self.codecs[n].decode(value[n]) for n in range(len(self.codecs)))
        )


class TupleJSONCodec(_TupleCodec[PT, JSONType], JSONCodec[PT]):
    """..."""

    def __init__(self, python_type: Any):
        _TupleCodec.__init__(self, python_type, JSONCodec)
        JSONCodec.__init__(self, python_type)

    def encode(self, value: PT) -> JSONType:
        return self._encode(value)

    def decode(self, value: JSONType) -> PT:
        return self._decode(value)


class TupleStringCodec(_TupleCodec[PT, StringType], StringCodec[PT]):
    """..."""

    def __init__(self, python_type: Any):
        _TupleCodec.__init__(self, python_type, StringCodec)
        StringCodec.__init__(self, python_type)

    def encode(self, value: PT) -> StringType:
        return _csv_encode(self._encode(value))

    def decode(self, value: StringType) -> PT:
        return self._decode(_csv_decode(value))


class TupleBinaryCodec(BinaryCodec[PT]):
    """..."""

    content_type = APPLICATION_JSON

    @staticmethod
    def handles(python_type: Any) -> bool:
        return TupleJSONCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = TupleJSONCodec(python_type)

    def encode(self, value: PT) -> BinaryType:
        return json.dumps(self.codec.encode(value)).encode()

    def decode(self, value: BinaryType) -> PT:
        return self.codec.decode(_s2j(_b2s(value)))


# ----- Mapping -----


class MappingJSONCodec(JSONCodec[PT]):
    """..."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        origin = get_origin(python_type) or python_type
        return is_subclass(origin, Mapping) and not getattr(origin, "__annotations__", None)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        python_type = strip_annotations(python_type)
        args = get_args(python_type) or (Any, Any)
        if len(args) != 2:
            raise TypeError("expecting Mapping[KT, VT]")
        self.key_codec = StringCodec.get(args[0])
        self.value_codec = JSONCodec.get(args[1])

    def encode(self, value: PT) -> JSONType:
        if not isinstance(value, Mapping):
            raise EncodeError
        result = {}
        for k, v in value.items():
            key = self.key_codec.encode(k)
            with CodecError.path_on_error(key):
                result[key] = self.value_codec.encode(v)
        return result

    def decode(self, value: JSONType) -> PT:
        if not isinstance(value, Mapping):
            raise DecodeError
        result = {}
        for k, v in value.items():
            key = self.key_codec.decode(k)
            with CodecError.path_on_error(key):
                result[key] = self.value_codec.decode(v)
        return result


class MappingStringCodec(StringCodec[PT]):
    """..."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        return MappingJSONCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = MappingJSONCodec(python_type)

    def encode(self, value: PT) -> StringType:
        return json.dumps(self.codec.encode(value))

    def decode(self, value: StringType) -> PT:
        return self.codec.decode(_s2j(value))


class MappingBinaryCodec(BinaryCodec[PT]):
    """..."""

    content_type = APPLICATION_JSON

    @staticmethod
    def handles(python_type: Any) -> bool:
        return MappingStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = MappingStringCodec(python_type)

    def encode(self, value: PT) -> BinaryType:
        if not isinstance(value, Mapping):
            raise EncodeError
        return self.codec.encode(value).encode()

    def decode(self, value: BinaryType) -> PT:
        return self.codec.decode(_b2s(value))


# ----- Iterable -----


class _IterableCodec(Codec[PT, TT]):
    """..."""

    _AVOID = str | bytes | bytearray | Mapping | tuple

    @classmethod
    def handles(cls, python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        origin = get_origin(python_type) or python_type
        return is_subclass(origin, Iterable) and not is_subclass(origin, _IterableCodec._AVOID)

    def __init__(self, python_type: Any, base_codec_type: type[Codec[PT, TT]]):
        python_type = strip_annotations(python_type)
        origin = get_origin(python_type) or python_type
        args = get_args(python_type) or (Any,)
        if len(args) != 1:
            raise TypeError("expecting Iterable[T]")
        self.decode_type = list if origin is Iterable else python_type
        self.codec = base_codec_type.get(args[0])
        self.is_set = is_subclass(origin, Set)

    def _encode(self, value: PT) -> list[Any]:
        if not isinstance(value, Iterable) or isinstance(value, _IterableCodec._AVOID):
            raise EncodeError
        if self.is_set:
            value = sorted(value)
        # TODO: path
        return [self.codec.encode(item) for item in value]

    def _decode(self, value: list[Any]) -> PT:
        if not isinstance(value, list):
            raise DecodeError
        # TODO: path
        return self.decode_type((self.codec.decode(item) for item in value))


class IterableJSONCodec(_IterableCodec[PT, JSONType], JSONCodec[PT]):
    """..."""

    def __init__(self, python_type: Any):
        JSONCodec.__init__(self, python_type)
        _IterableCodec.__init__(self, python_type, JSONCodec)

    def encode(self, value: PT) -> JSONType:
        return self._encode(value)

    def decode(self, value: JSONType) -> PT:
        return self._decode(value)


class IterableStringCodec(_IterableCodec[PT, StringType], StringCodec[PT]):
    """..."""

    def __init__(self, python_type: Any):
        JSONCodec.__init__(self, python_type)
        _IterableCodec.__init__(self, python_type, StringCodec)

    def encode(self, value: PT) -> StringType:
        return _csv_encode(self._encode(value))

    def decode(self, value: StringType) -> PT:
        return self._decode(_csv_decode(value))


class IterableBinaryCodec(BinaryCodec[PT]):
    """..."""

    content_type = APPLICATION_JSON

    @staticmethod
    def handles(python_type: Any) -> bool:
        return IterableJSONCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = IterableJSONCodec(python_type)

    def encode(self, value: PT) -> BinaryType:
        return json.dumps(self.codec.encode(value)).encode()

    def decode(self, value: BinaryType) -> PT:
        return self.codec.decode(_s2j(_b2s(value)))


# ----- Generic -----


class _GenericCodec(Generic[PT, TT]):
    """..."""

    @classmethod
    def handles(cls, python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        with suppress(AttributeError):
            return Generic in get_origin(python_type).__bases__
        return False

    def __init__(self, python_type: Any, base_codec_type: type[Codec[PT, TT]]):
        self.raw_type = strip_annotations(python_type)
        self.base_codec_type = base_codec_type

    @property
    def _codec(self) -> Codec[PT, TT]:
        return self.base_codec_type.get(get_origin(self.raw_type))

    def encode(self, value: PT) -> TT:
        with fondat.types.capture_typevars(self.raw_type):
            return self._codec.encode(value)

    def decode(self, value: PT) -> TT:
        with fondat.types.capture_typevars(self.raw_type):
            return self._codec.decode(value)


class GenericJSONCodec(_GenericCodec[PT, JSONType], JSONCodec[PT]):
    def __init__(self, python_type: Any):
        JSONCodec.__init__(self, python_type)
        _GenericCodec.__init__(self, python_type, JSONCodec)


class GenericStringCodec(_GenericCodec[PT, StringType], StringCodec[PT]):
    def __init__(self, python_type: Any):
        StringCodec.__init__(self, python_type)
        _GenericCodec.__init__(self, python_type, StringCodec)


class GenericBinaryCodec(_GenericCodec[PT, BinaryType], BinaryCodec[PT]):
    def __init__(self, python_type: Any):
        BinaryCodec.__init__(self, python_type)
        _GenericCodec.__init__(self, python_type, BinaryCodec)


# ----- TypeVar -----


class _TypeVarCodec(Generic[PT, TT]):
    """..."""

    _cache = False  # TypeVars can be reused

    @classmethod
    def handles(cls, python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return isinstance(python_type, TypeVar)

    def __init__(self, python_type: Any, base_codec_type: type[Codec[PT, TT]]):
        self.raw_type = strip_annotations(python_type)
        self.base_codec_type = base_codec_type

    @property
    def _codec(self) -> Codec[PT, TT]:
        resolved = fondat.types.resolve_typevar(self.raw_type)
        return self.base_codec_type.get(resolved)

    def encode(self, value: PT) -> TT:
        return self._codec.encode(value)

    def decode(self, value: TT) -> PT:
        return self._codec.decode(value)


class TypeVarJSONCodec(_TypeVarCodec[PT, JSONType], JSONCodec[PT]):
    """..."""

    def __init__(self, python_type: Any):
        JSONCodec.__init__(self, python_type)
        _TypeVarCodec.__init__(self, python_type, JSONCodec)


class TypeVarStringCodec(_TypeVarCodec[PT, StringType], StringCodec[PT]):
    """..."""

    def __init__(self, python_type: Any):
        StringCodec.__init__(self, python_type)
        _TypeVarCodec.__init__(self, python_type, StringCodec)


class TypeVarBinaryCodec(_TypeVarCodec[PT, BinaryType], BinaryCodec[PT]):
    """..."""

    def __init__(self, python_type: Any):
        BinaryCodec.__init__(self, python_type)
        _TypeVarCodec.__init__(self, python_type, BinaryCodec)
        self.content_type = self.codec.content_type


# ----- dataclass -----


class DataclassJSONCodec(JSONCodec[PT]):

    # keywords have _ suffix in dataclass fields (e.g. "in_", "for_", ...)
    _dc_kw = {k + "_": k for k in keyword.kwlist}

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return dataclasses.is_dataclass(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.raw_type = strip_annotations(python_type)
        self.hints = typing.get_type_hints(self.raw_type, include_extras=True)

    @property
    def _codecs(self) -> Iterable[JSONCodec[Any]]:
        return {key: JSONCodec.get(hint) for key, hint in self.hints.items()}

    def encode(self, value: PT) -> JSONType:
        if not isinstance(value, self.raw_type):
            raise EncodeError
        result = {}
        for field in dataclasses.fields(self.raw_type):
            v = getattr(value, field.name, None)
            if v is not None:
                with CodecError.path_on_error(field.name):
                    result[
                        DataclassJSONCodec._dc_kw.get(field.name, field.name)
                    ] = self._codecs[field.name].encode(v)
        return result

    def decode(self, value: JSONType) -> PT:
        if not isinstance(value, dict):
            raise DecodeError
        kwargs = {}
        for field in dataclasses.fields(self.raw_type):
            try:
                with CodecError.path_on_error(field.name):
                    kwargs[field.name] = self._codecs[field.name].decode(
                        value[DataclassJSONCodec._dc_kw.get(field.name, field.name)]
                    )
            except KeyError:
                if (
                    is_optional(field.type)
                    and field.default is dataclasses.MISSING
                    and field.default_factory is dataclasses.MISSING
                ):
                    kwargs[field.name] = None
        with _wrap(DecodeError):
            return self.raw_type(**kwargs)


class DataclassStringCodec(StringCodec[PT]):
    """..."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        return DataclassJSONCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = DataclassJSONCodec(python_type)

    def encode(self, value: PT) -> StringType:
        return json.dumps(self.codec.encode(value))

    def decode(self, value: StringType) -> PT:
        return self.codec.decode(_s2j(value))


class DataclassBinaryCodec(BinaryCodec[PT]):
    """..."""

    content_type = APPLICATION_JSON

    @staticmethod
    def handles(python_type: Any) -> bool:
        return DataclassStringCodec.handles(python_type)

    def __init__(self, python_type: Any):
        super().__init__(python_type)
        self.codec = DataclassStringCodec(python_type)

    def encode(self, value: PT) -> BinaryType:
        return self.codec.encode(value).encode()

    def decode(self, value: BinaryType) -> PT:
        return self.codec.decode(_b2s(value))


# ----- UnionType/Union -----


class _UnionCodec(Codec[PT, TT]):
    """..."""

    @classmethod
    def handles(cls, python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return get_origin(python_type) in {UnionType, Union}

    def __init__(self, python_type: Any, base_codec_type: type[Codec[PT, TT]]):
        python_type = strip_annotations(python_type)
        self.codecs = tuple(base_codec_type.get(type) for type in set(get_args(python_type)))

    def encode(self, value: PT) -> TT:
        for codec in self.codecs:
            with suppress(EncodeError):
                return codec.encode(value)
        raise EncodeError

    def decode(self, value: TT) -> PT:
        for codec in self.codecs:
            with suppress(DecodeError):
                return codec.decode(value)
        raise DecodeError


class UnionJSONCodec(_UnionCodec[PT, Any], JSONCodec[PT]):
    """..."""

    def __init__(self, python_type: Any):
        JSONCodec.__init__(self, python_type)
        _UnionCodec.__init__(self, python_type, JSONCodec)


class UnionStringCodec(_UnionCodec[PT, str], StringCodec[PT]):
    """..."""

    def __init__(self, python_type: Any):
        StringCodec.__init__(self, python_type)
        _UnionCodec.__init__(self, python_type, StringCodec)


class UnionBinaryCodec(_UnionCodec[PT, bytes | bytearray], BinaryCodec[PT]):
    """..."""

    content_type = APPLICATION_OCTET_STREAM

    @staticmethod
    def handles(python_type: Any) -> bool:
        return _UnionCodec.handles(python_type)

    def __init__(self, python_type: Any):
        BinaryCodec.__init__(self, python_type)
        _UnionCodec.__init__(self, python_type, BinaryCodec)


# ----- Literal -----


_VT = namedtuple("VT", "value,type")


class _LiteralCodec(Codec[PT, TT]):
    """..."""

    @classmethod
    def handles(cls, python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return get_origin(python_type) is Literal

    def __init__(self, python_type: Any, base_codec_type: type[Codec[PT, TT]]):
        python_type = strip_annotations(python_type)
        self.vts = {_VT(v, type(v)) for v in fondat.types.literal_values(python_type)}
        self.codecs = {vt: base_codec_type.get(vt.type) for vt in self.vts}

    def encode(self, value: PT) -> TT:
        with _wrap(EncodeError):
            return self.codecs[_VT(value, type(value))].encode(value)
        raise EncodeError

    def decode(self, value: TT) -> PT:
        for codec in self.codecs.values():
            with suppress(DecodeError):
                decoded = codec.decode(value)
                if _VT(decoded, type(decoded)) in self.vts:
                    return decoded
        raise DecodeError


class LiteralStringCodec(_LiteralCodec[PT, StringType], StringCodec[PT]):
    """..."""

    def __init__(self, python_type: Any):
        StringCodec.__init__(self, python_type)
        _LiteralCodec.__init__(self, python_type, StringCodec)


class LiteralBinaryCodec(_LiteralCodec[PT, BinaryType], BinaryCodec[PT]):
    """..."""

    content_type = APPLICATION_OCTET_STREAM

    def __init__(self, python_type: Any):
        StringCodec.__init__(self, python_type)
        _LiteralCodec.__init__(self, python_type, BinaryCodec)


class LiteralJSONCodec(_LiteralCodec[PT, JSONType], JSONCodec[PT]):
    """..."""

    def __init__(self, python_type: Any):
        JSONCodec.__init__(self, python_type)
        _LiteralCodec.__init__(self, python_type, JSONCodec)


# ----- Any -----


class AnyStringCodec(StringCodec[Any]):
    """String codec for Any."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return python_type is Any

    def encode(self, value: Any) -> StringType:
        return StringCodec.get(type(value)).encode(value)

    def decode(self, value: StringType) -> Any:
        return value


class AnyBinaryCodec(BinaryCodec[Any]):
    """Binary codec for Any."""

    content_type = APPLICATION_OCTET_STREAM

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return python_type is Any

    def encode(self, value: Any) -> BinaryType:
        return BinaryCodec.get(type(value)).encode(value)

    def decode(self, value: BinaryType) -> Any:
        if not isinstance(value, (bytes, bytearray)):
            raise DecodeError
        return value


class AnyJSONCodec(JSONCodec[Any]):
    """JSON codec for Any."""

    @staticmethod
    def handles(python_type: Any) -> bool:
        python_type = strip_annotations(python_type)
        return python_type is Any

    def encode(self, value: Any) -> Any:
        return JSONCodec.get(type(value)).encode(value)

    def decode(self, value: Any) -> Any:
        return value



"""
Module to implement resources composed of operations.

A resource is an addressible object that exposes operations through a uniform interface. A
resource class contains operation methods, each decorated with the @operation decorator.

A resource can expose subordinate resources through an attribute, method or property that is
also decorated as a resource.

For information on how security policies are evaluated, see the `authorize` function.
"""

import asyncio
import fondat.context as context
import fondat.monitor as monitor
import functools
import hashlib
import inspect
import json
import logging
import types
import wrapt

from collections.abc import Iterable, Mapping
from contextlib import contextmanager
from copy import deepcopy
from fondat.codec import JSONCodec
from fondat.error import BadRequestError, ForbiddenError, UnauthorizedError
from fondat.security import Policy
from fondat.types import literal_values
from fondat.validation import ValidationError, validate_arguments
from typing import Any, Literal, TypeVar


T = TypeVar("T")


_logger = logging.getLogger(__name__)


def _summary(function):
    """
    Derive summary information from a function's docstring or name. The summary is the first
    sentence of the docstring, ending in a period, or if no dostring is present, the
    function's name capitalized.
    """
    if not function.__doc__:
        return f"{function.__name__.capitalize()}."
    result = []
    for word in function.__doc__.split():
        result.append(word)
        if word.endswith("."):
            break
    return " ".join(result)


def _hash_json(key: Any) -> bytes:
    """
    Return a deterministic, unique hash value for a given JSON object model value.
    """
    return hashlib.sha256(json.dumps(key, sort_keys=True).encode()).digest()


async def authorize(policies: Iterable[Policy]):
    """
    Evaluate the specified security policies.

    Parameters:
    • policies: security policies to evaluate

    Security exceptions are: UnauthorizedError and ForbiddenError.

    This function evaluates the security policies in the order specified. If a non-security
    exception occurs, it shall be immediately raised, ceasing further evaluaton. Otherwise,
    if one of the security policies does not raise a security exception, then authorization
    shall be immediately granted, ceasing further evaluation.

    If security policies all raise security exceptions, then the first ForbiddenError exception
    is raised if encountered, otherwise the first UnauthorizedError exception is raised.
    """
    exception = None
    for policy in policies or ():
        try:
            await policy.apply()
            return  # security policy authorized the operation
        except ForbiddenError as fe:
            if not isinstance(exception, ForbiddenError):
                exception = fe
        except UnauthorizedError as ue:
            if not exception:
                exception = ue
        except:
            raise
    if exception:
        raise exception


def resource(wrapped: type[T] | None = None, *, tag: str | None = None) -> type[T]:
    """
    Decorate a class as a resource containing operations and/or subordinate resources.

    Parameters:
    • tag: tag to group resources  [resource class name]
    """

    if wrapped is None:
        return functools.partial(resource, tag=tag)
    wrapped._fondat_resource = types.SimpleNamespace(tag=tag or wrapped.__name__)
    return wrapped


Method = Literal["get", "put", "post", "delete", "patch"]

_methods = literal_values(Method)


@contextmanager
def _suppress_and_log():
    try:
        yield
    except Exception as exception:
        _logger.debug(exception, exc_info=True)


@validate_arguments
def operation(
    wrapped=None,
    *,
    method: Method | None = None,
    type: Literal["query", "mutation"] | None = None,
    policies: Iterable[Policy] | None = None,
    publish: bool = True,
    deprecated: bool = False,
    cache: Any | None = None,
):
    """
    Decorate a resource coroutine as an operation.

    Parameters:
    • method: method of operation  [inferred from wrapped coroutine name]
    • type: type of operation  [inferred from method]
    • policies: security policies for the operation
    • publish: publish the operation in documentation
    • deprecated: flag the operation as deprecated
    • cache: resource to cache operation results

    When an operation is called:
    • its arguments are copied to prevent side effects
    • its arguments are validated against their type hints

    Exceptions that an operation should raise:
    • fondat.error.Error (e.g. BadRequestError)
    • fondat.validation.ValidationError (automatically reraised as a BadRequestError)

    The cache is a resource that exposes cache entry resources. A MemoryResource can serve as
    an operation cache resource out of the box. It is safe for a single cache resource to be
    shared by multiple operations.
    """

    if wrapped is None:
        return functools.partial(
            operation,
            method=method,
            type=type,
            policies=policies,
            publish=publish,
            deprecated=deprecated,
            cache=cache,
        )

    if not asyncio.iscoroutinefunction(wrapped):
        raise TypeError("operation must be a coroutine")

    if method is None:
        method = wrapped.__name__
        if method not in _methods:
            raise TypeError(f"method must be one of: {_methods}")

    if not type:
        type = "query" if method == "get" else "mutation"

    description = wrapped.__doc__ or wrapped.__name__
    summary = _summary(wrapped)

    sig = inspect.signature(wrapped)
    params = [param for param in sig.parameters.values()]
    returns = sig.return_annotation if sig.return_annotation is not sig.empty else None

    if not params or params[0].name != "self":
        raise TypeError("operation first parameter must be self")
    for param in params[1:]:
        if param.kind is param.VAR_POSITIONAL:
            raise TypeError("operation with *args is not supported")
        if param.kind is param.VAR_KEYWORD:
            raise TypeError("operation with **kwargs is not supported")
        if param.annotation is param.empty:
            raise TypeError(f"operation parameter must have type hint: {param.name}")

    if cache and not is_resource(cache):
        raise TypeError("cache must be a resource")

    @wrapt.decorator
    async def wrapper(wrapped, instance, args, kwargs):
        args, kwargs = deepcopy(args), deepcopy(kwargs)  # avoid side effects
        resource_name = f"{instance.__class__.__module__}.{instance.__class__.__qualname__}"
        operation = getattr(wrapped, "_fondat_operation")
        arguments = dict(zip((p.name for p in params[1:]), args)) | kwargs
        operation_name = wrapped.__name__
        tags = {"resource": resource_name, "operation": operation_name}
        _logger.debug("operation: %s.%s(%s)", resource_name, operation_name, arguments)
        with context.push(tags | {"context": "fondat.operation", "arguments": arguments}):
            async with monitor.counter(
                name="operation_invocations", tags=tags, status="status"
            ):
                async with monitor.timer(name="operation_duration", tags=tags):
                    await authorize(operation.policies)
                    if cache:
                        with _suppress_and_log():
                            cache_entry = cache[
                                _hash_json(
                                    tags | {"arguments": JSONCodec.get(Any).encode(arguments)}
                                )
                            ]
                            return JSONCodec.get(returns).decode(await cache_entry.get())
                    try:
                        result = await wrapped(*args, **kwargs)
                    except ValidationError as ve:
                        raise BadRequestError from ve
                    if cache:
                        with _suppress_and_log():
                            await cache_entry.put(JSONCodec.get(returns).encode(result))
                    return result

    wrapped._fondat_operation = types.SimpleNamespace(
        method=method,
        type=type,
        policies=policies,
        publish=publish,
        deprecated=deprecated,
        summary=summary,
        description=description,
    )

    wrapped = validate_arguments(wrapped)
    return wrapper(wrapped)


def query(wrapped=None, *, method: str = "get", **kwargs):
    """Decorator to define a query operation."""

    if wrapped is None:
        return functools.partial(
            query,
            method=method,
            **kwargs,
        )

    return operation(wrapped, type="query", method=method, **kwargs)


def mutation(wrapped=None, *, method: str = "post", **kwargs):
    """Decorator to define an mutation operation."""

    if wrapped is None:
        return functools.partial(
            mutation,
            method=method,
            **kwargs,
        )

    return operation(wrapped, type="mutation", method=method, **kwargs)


def container_resource(resources: Mapping[str, Any], tag: str | None = None):
    """
    Create a resource to contain subordinate resources.

    Parameters:
    • resources: mapping of resource names to resource objects
    • tag: tag to group the resource

    Suborindates are accessed as attributes by name.
    """

    @resource(tag=tag)
    class Container:
        def __getattr__(self, name):
            try:
                return resources[name]
            except KeyError:
                raise AttributeError(f"no such resource: {name}")

        def __dir__(self):
            return [*super().__dir__(), *resources.keys()]

    return Container()


def is_resource(obj_or_type: Any) -> bool:
    """Return if object or type is a resource."""
    return hasattr(obj_or_type, "_fondat_resource")


def is_operation(obj_or_type: Any) -> bool:
    """Return if object is a resource operation method."""
    return hasattr(obj_or_type, "_fondat_operation")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fondat']

package_data = \
{'': ['*']}

install_requires = \
['aiosqlite>=0.17,<0.18',
 'iso8601>=1.1,<2.0',
 'multidict>=6.0,<7.0',
 'wrapt>=1.14,<2.0']

setup_kwargs = {
    'name': 'fondat',
    'version': '4.0.4',
    'description': 'A foundation for resource-oriented backend applications.',
    'long_description': '# fondat\n\n[![PyPI](https://img.shields.io/pypi/v/fondat)](https://pypi.org/project/fondat/)\n[![Python](https://img.shields.io/pypi/pyversions/fondat)](https://python.org/)\n[![GitHub](https://img.shields.io/badge/github-main-blue.svg)](https://github.com/fondat/fondat/)\n[![Test](https://github.com/fondat/fondat/workflows/test/badge.svg)](https://github.com/fondat/fondat/actions?query=workflow/test)\n[![License](https://img.shields.io/github/license/fondat/fondat.svg)](https://github.com/fondat/fondat/blob/main/LICENSE)\n\nA foundation for resource-oriented backend applications.\n\n## Introduction\n\nFondat is a foundation for resource-oriented backend applications in Python.\n\n## Install\n\n```\npip install fondat\n```\n\n## Develop\n\n```\npoetry install\npoetry run pre-commit install\n```\n\n## Test\n\n```\npoetry run pytest\n```',
    'author': 'fondat authors',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/fondat/fondat/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""
Fondat monitor module.

A global "monitors" object is a list of monitors and a monitor itself. Applications are free
to add/remove their own monitors to/from this object.
"""

import asyncio
import time

from contextlib import asynccontextmanager
from dataclasses import field
from datetime import datetime, timezone
from fondat.data import datacls
from fondat.validation import MinLen, validate
from types import NoneType
from typing import Annotated, Literal


# type aliases
Name = Annotated[str, MinLen(1)]
Type = Literal["counter", "gauge"]
Tags = dict[str, str]
Value = int | float


_now = lambda: datetime.now(tz=timezone.utc)


@datacls
class Measurement:
    """
    An individual measurement.

    Parameters and attributes:
    • name: name of the measurement
    • tags: key-value pairs that qualify the measurement
    • timestamp: date and time of the measurement  [now]
    • type: type of measurement
    • value: measured value
    • unit: unit of measure

    Name should be an identifier, expressed in snake_case, which describes the metric being
    measured (e.g. "operation_invocations", "operation_duration").

    Unit should be a standard symbol (e.g. SI symbol "s" for seconds). If unit is a rate,
    symbols separated by a slash "/" character should be used (e.g. "m/s" for metres per
    second).
    """

    name: Name
    tags: Tags | None
    timestamp: datetime = field(default_factory=_now)
    type: Type
    value: Value
    unit: str | None

    def __post_init__(self):
        validate(self, Measurement)


class Monitor:
    """Base class for a monitor that records measurements."""

    async def record(self, measurement: Measurement) -> NoneType:
        """Record a measurement."""
        raise NotImplementedError

    async def flush(self) -> NoneType:
        """Flush all cached measurements. Base class implementation does nothing."""
        return


class Monitors(Monitor, list[Monitor]):
    """A list of monitors, to which all measurements are recorded."""

    async def record(self, measurement: Measurement):
        """Record a measurement in monitors."""
        await asyncio.gather(monitor.record(measurement) for monitor in self)

    async def flush(self):
        """Flush all cached measurements."""
        await asyncio.gather(monitor.flush() for monitor in self)


monitors = Monitors()


@asynccontextmanager
async def timer(
    *,
    name: Name,
    tags: Tags | None = None,
    monitor: Monitor | None = None,
):
    """
    An asynchronous context manager that times the execution of work and records it as a
    gauge measurement of duration in seconds.

    Parameters:
    • name: measurement name
    • tags: key-value pairs that qualify the measurement
    • monitor: monitor to record measurement  [global monitors]

    If an exception is raised during execution, the measurement will not be recorded.
    """
    begin = time.perf_counter()
    yield
    duration = time.perf_counter() - begin
    await record(
        Measurement(name=name, type="gauge", value=duration, unit="s", tags=tags),
        monitor,
    )


@asynccontextmanager
async def counter(
    *,
    name: Name,
    tags: Tags | None = None,
    monitor: Monitor | None = None,
    status: str | None = None,
):
    """
    An asynchronous context manager that counts the number of executions, and optional status.

    Parameters:
    • name: measurement name
    • tags: key-value pairs that qualify the measurement
    • monitor: monitor to record measurement  [global monitors]
    • status: tag to record the status of execution

    If recording status, the value "success" will be added as a tag if execution was
    successful, or "failure" if an exception was raised during execution.
    """
    exception = None
    try:
        yield
    except Exception as e:
        exception = e
    if status:
        tags = {**(tags or {}), "status": "success" if not exception else "failure"}
    await record(
        Measurement(name=name, type="counter", value=1, tags=tags),
        monitor,
    )
    if exception:
        raise exception


async def record(measurement: Measurement, monitor: Monitor | None):
    """
    Record a measurement.

    Parameters:
    • measurement: measurement to record
    • monitor: monitor to record measurement  [global monitors]
    """
    await (monitor if monitor is not None else monitors).record(measurement)
